__author__ = "sibirrer"

import numpy.testing as npt
import numpy as np
import pytest

import lenstronomy.Util.param_util as param_util
from lenstronomy.LensModel.lens_model import LensModel
from lenstronomy.LightModel.light_model import LightModel
from lenstronomy.PointSource.point_source import PointSource
from lenstronomy.ImSim.image_model import ImageModel
from lenstronomy.ImSim.image_linear_solve import ImageLinearFit
import lenstronomy.Util.simulation_util as sim_util
from lenstronomy.LensModel.Solver.lens_equation_solver import LensEquationSolver
from lenstronomy.Data.imaging_data import ImageData
from lenstronomy.Data.psf import PSF
from lenstronomy.ImSim.differential_extinction import DifferentialExtinction
from lenstronomy.Util import util


class TestImageModel(object):
    """Tests the source model routines."""

    def setup_method(self):
        # data specifics
        sigma_bkg = 0.05  # background noise per pixel
        exp_time = 100  # exposure time (arbitrary units, flux per pixel is in units #photons/exp_time unit)
        numPix = 100  # cutout pixel size
        deltaPix = 0.05  # pixel size in arcsec (area per pixel = deltaPix**2)
        fwhm = 0.5  # full width half max of PSF

        # PSF specification

        kwargs_data = sim_util.data_configure_simple(
            numPix, deltaPix, exp_time, sigma_bkg, inverse=True
        )
        data_class = ImageData(**kwargs_data)
        kwargs_psf = {
            "psf_type": "GAUSSIAN",
            "fwhm": fwhm,
            "truncation": 5,
            "pixel_size": deltaPix,
        }
        psf_class = PSF(**kwargs_psf)
        kernel = psf_class.kernel_point_source
        kwargs_psf = {
            "psf_type": "PIXEL",
            "kernel_point_source": kernel,
            "psf_error_map": np.ones_like(kernel) * 0.001 * kernel**2,
        }
        psf_class = PSF(**kwargs_psf)

        # 'EXERNAL_SHEAR': external shear
        kwargs_shear = {
            "gamma1": 0.01,
            "gamma2": 0.01,
        }  # gamma_ext: shear strength, psi_ext: shear angel (in radian)
        phi, q = 0.2, 0.8
        e1, e2 = param_util.phi_q2_ellipticity(phi, q)
        kwargs_spemd = {
            "theta_E": 1.0,
            "gamma": 1.8,
            "center_x": 0,
            "center_y": 0,
            "e1": e1,
            "e2": e2,
        }

        lens_model_list = ["SPEP", "SHEAR"]
        self.kwargs_lens = [kwargs_spemd, kwargs_shear]
        lens_model_class = LensModel(lens_model_list=lens_model_list)
        # list of light profiles (for lens and source)
        # 'SERSIC': spherical Sersic profile
        kwargs_sersic = {
            "amp": 1.0,
            "R_sersic": 0.1,
            "n_sersic": 2,
            "center_x": 0,
            "center_y": 0,
        }
        # 'SERSIC_ELLIPSE': elliptical Sersic profile
        phi, q = 0.2, 0.9
        e1, e2 = param_util.phi_q2_ellipticity(phi, q)
        kwargs_sersic_ellipse = {
            "amp": 1.0,
            "R_sersic": 0.6,
            "n_sersic": 7,
            "center_x": 0,
            "center_y": 0,
            "e1": e1,
            "e2": e2,
        }

        lens_light_model_list = ["SERSIC"]
        self.kwargs_lens_light = [kwargs_sersic]
        lens_light_model_class = LightModel(light_model_list=lens_light_model_list)
        source_model_list = ["SERSIC_ELLIPSE"]
        self.kwargs_source = [kwargs_sersic_ellipse]
        source_model_class = LightModel(light_model_list=source_model_list)
        self.kwargs_ps = [
            {"ra_source": 0.01, "dec_source": 0.0, "source_amp": 1.0}
        ]  # quasar point source position in the source plane and intrinsic brightness
        point_source_class = PointSource(
            point_source_type_list=["SOURCE_POSITION"], fixed_magnification_list=[True]
        )
        kwargs_numerics = {
            "supersampling_factor": 2,
            "supersampling_convolution": False,
        }
        imageModel = ImageModel(
            data_class,
            psf_class,
            lens_model_class,
            source_model_class,
            lens_light_model_class,
            point_source_class,
            kwargs_numerics=kwargs_numerics,
        )
        image_sim = sim_util.simulate_simple(
            imageModel,
            self.kwargs_lens,
            self.kwargs_source,
            self.kwargs_lens_light,
            self.kwargs_ps,
        )
        data_class.update_data(image_sim)

        self.imageModel = ImageLinearFit(
            data_class,
            psf_class,
            lens_model_class,
            source_model_class,
            lens_light_model_class,
            point_source_class,
            kwargs_numerics=kwargs_numerics,
        )
        self.solver = LensEquationSolver(lensModel=self.imageModel.LensModel)

    def test_source_surface_brightness(self):
        source_model = self.imageModel.source_surface_brightness(
            self.kwargs_source, self.kwargs_lens, unconvolved=False, de_lensed=True
        )
        assert len(source_model) == 100

        source_model = self.imageModel.source_surface_brightness(
            self.kwargs_source, self.kwargs_lens, unconvolved=False, de_lensed=False
        )
        assert len(source_model) == 100
        npt.assert_almost_equal(
            source_model[10, 10], 0.13939841209844345 * 0.05**2, decimal=4
        )

        source_model = self.imageModel.source_surface_brightness(
            self.kwargs_source, self.kwargs_lens, unconvolved=True, de_lensed=False
        )
        assert len(source_model) == 100
        npt.assert_almost_equal(
            source_model[10, 10], 0.13536114618182182 * 0.05**2, decimal=4
        )

    def test_lens_surface_brightness(self):
        lens_flux = self.imageModel.lens_surface_brightness(
            self.kwargs_lens_light, unconvolved=False
        )
        print(np.sum(lens_flux), "test lens flux")
        npt.assert_almost_equal(lens_flux[50, 50], 0.0010788981265391802, decimal=4)
        # npt.assert_almost_equal(lens_flux[50, 50], 0.54214440654021534 * 0.05 ** 2, decimal=4)

        lens_flux = self.imageModel.lens_surface_brightness(
            self.kwargs_lens_light, unconvolved=True
        )
        npt.assert_almost_equal(
            lens_flux[50, 50], 4.7310552067454452 * 0.05**2, decimal=4
        )

    def test_image_linear_solve(self):
        model, error_map, cov_param, param = self.imageModel.image_linear_solve(
            self.kwargs_lens,
            self.kwargs_source,
            self.kwargs_lens_light,
            self.kwargs_ps,
            inv_bool=False,
        )
        chi2_reduced = self.imageModel.reduced_chi2(model, error_map)
        npt.assert_almost_equal(chi2_reduced, 1, decimal=1)

    def test_linear_response_matrix(self):
        A = self.imageModel.linear_response_matrix(
            self.kwargs_lens, self.kwargs_source, self.kwargs_lens_light, self.kwargs_ps
        )
        n, m = np.shape(A)
        assert n == 3
        assert m == 100 * 100

    def test_image_with_params(self):
        model = self.imageModel.image(
            self.kwargs_lens,
            self.kwargs_source,
            self.kwargs_lens_light,
            self.kwargs_ps,
            unconvolved=False,
            source_add=True,
            lens_light_add=True,
            point_source_add=True,
        )
        error_map = self.imageModel._error_map_psf(self.kwargs_lens, self.kwargs_ps)
        chi2_reduced = self.imageModel.reduced_chi2(model, error_map)
        npt.assert_almost_equal(chi2_reduced, 1, decimal=1)

    def test_likelihood_data_given_model(self):
        logL = self.imageModel.likelihood_data_given_model(
            self.kwargs_lens,
            self.kwargs_source,
            self.kwargs_lens_light,
            self.kwargs_ps,
            source_marg=False,
        )
        npt.assert_almost_equal(logL, -5000, decimal=-3)

        logLmarg = self.imageModel.likelihood_data_given_model(
            self.kwargs_lens,
            self.kwargs_source,
            self.kwargs_lens_light,
            self.kwargs_ps,
            source_marg=True,
        )
        npt.assert_almost_equal(logL - logLmarg, 0, decimal=-3)
        assert logLmarg < logL

    def test_reduced_residuals(self):
        model = sim_util.simulate_simple(
            self.imageModel,
            self.kwargs_lens,
            self.kwargs_source,
            self.kwargs_lens_light,
            self.kwargs_ps,
            no_noise=True,
        )
        residuals = self.imageModel.reduced_residuals(model, error_map=0)
        npt.assert_almost_equal(np.std(residuals), 1.01, decimal=1)

        chi2 = self.imageModel.reduced_chi2(model, error_map=0)
        npt.assert_almost_equal(chi2, 1, decimal=1)

    def test_numData_evaluate(self):
        numData = self.imageModel.num_data_evaluate
        assert numData == 10000

    def test_num_param_linear(self):
        num_param_linear = self.imageModel.num_param_linear(
            self.kwargs_lens, self.kwargs_source, self.kwargs_lens_light, self.kwargs_ps
        )
        assert num_param_linear == 3

    def test_update_data(self):
        kwargs_data = sim_util.data_configure_simple(
            numPix=10, deltaPix=1, exposure_time=1, background_rms=1, inverse=True
        )
        data_class = ImageData(**kwargs_data)
        self.imageModel.update_data(data_class)
        assert self.imageModel.Data.num_pixel == 100

    def test_point_source_rendering(self):
        # initialize data

        numPix = 100
        deltaPix = 0.05
        kwargs_data = sim_util.data_configure_simple(
            numPix, deltaPix, exposure_time=1, background_rms=1
        )
        data_class = ImageData(**kwargs_data)
        kernel = np.zeros((5, 5))
        kernel[2, 2] = 1
        kwargs_psf = {
            "kernel_point_source": kernel,
            "psf_type": "PIXEL",
            "psf_error_map": np.ones_like(kernel) * 0.001,
        }
        psf_class = PSF(**kwargs_psf)
        lens_model_class = LensModel(["SPEP"])
        source_model_class = LightModel([])
        lens_light_model_class = LightModel([])
        kwargs_numerics = {
            "supersampling_factor": 2,
            "supersampling_convolution": True,
            "point_source_supersampling_factor": 1,
        }
        point_source_class = PointSource(
            point_source_type_list=["LENSED_POSITION"], fixed_magnification_list=[False]
        )
        makeImage = ImageModel(
            data_class,
            psf_class,
            lens_model_class,
            source_model_class,
            lens_light_model_class,
            point_source_class,
            kwargs_numerics=kwargs_numerics,
        )
        # chose point source positions
        x_pix = np.array([10, 5, 10, 90])
        y_pix = np.array([40, 50, 60, 50])
        ra_pos, dec_pos = makeImage.Data.map_pix2coord(x_pix, y_pix)
        e1, e2 = param_util.phi_q2_ellipticity(0, 0.8)
        kwargs_lens_init = [
            {"theta_E": 1, "gamma": 2, "e1": e1, "e2": e2, "center_x": 0, "center_y": 0}
        ]
        kwargs_else = [
            {
                "ra_image": ra_pos,
                "dec_image": dec_pos,
                "point_amp": np.ones_like(ra_pos),
            }
        ]
        image = makeImage.image(
            kwargs_lens_init,
            kwargs_source={},
            kwargs_lens_light={},
            kwargs_ps=kwargs_else,
        )
        # print(np.shape(model), 'test')
        # image = makeImage.ImageNumerics.array2image(model)
        for i in range(len(x_pix)):
            npt.assert_almost_equal(image[y_pix[i], x_pix[i]], 1, decimal=2)

        x_pix = np.array([10.5, 5.5, 10.5, 90.5])
        y_pix = np.array([40, 50, 60, 50])
        ra_pos, dec_pos = makeImage.Data.map_pix2coord(x_pix, y_pix)
        phi, q = 0.0, 0.8
        e1, e2 = param_util.phi_q2_ellipticity(phi, q)
        kwargs_lens_init = [
            {"theta_E": 1, "gamma": 2, "e1": e1, "e2": e2, "center_x": 0, "center_y": 0}
        ]
        kwargs_else = [
            {
                "ra_image": ra_pos,
                "dec_image": dec_pos,
                "point_amp": np.ones_like(ra_pos),
            }
        ]
        image = makeImage.image(
            kwargs_lens_init,
            kwargs_source={},
            kwargs_lens_light={},
            kwargs_ps=kwargs_else,
        )
        # image = makeImage.ImageNumerics.array2image(model)
        for i in range(len(x_pix)):
            print(int(y_pix[i]), int(x_pix[i] + 0.5))
            npt.assert_almost_equal(image[int(y_pix[i]), int(x_pix[i])], 0.5, decimal=1)
            npt.assert_almost_equal(
                image[int(y_pix[i]), int(x_pix[i] + 0.5)], 0.5, decimal=1
            )

    def test_point_source(self):
        pointSource = PointSource(
            point_source_type_list=["SOURCE_POSITION"], fixed_magnification_list=[True]
        )
        kwargs_ps = [{"source_amp": 1000, "ra_source": 0.1, "dec_source": 0.1}]
        lensModel = LensModel(lens_model_list=["SIS"])
        kwargs_lens = [{"theta_E": 1, "center_x": 0, "center_y": 0}]
        numPix = 64
        deltaPix = 0.13
        kwargs_data = sim_util.data_configure_simple(
            numPix, deltaPix, exposure_time=1, background_rms=1
        )
        data_class = ImageData(**kwargs_data)

        psf_type = "GAUSSIAN"
        fwhm = 0.9
        kwargs_psf = {"psf_type": psf_type, "fwhm": fwhm}
        psf_class = PSF(**kwargs_psf)
        imageModel = ImageModel(
            data_class=data_class,
            psf_class=psf_class,
            lens_model_class=lensModel,
            point_source_class=pointSource,
        )
        image = imageModel.image(kwargs_lens=kwargs_lens, kwargs_ps=kwargs_ps)
        assert np.sum(image) > 0

    def test_error_map_source(self):
        sourceModel = LightModel(light_model_list=["UNIFORM", "UNIFORM"])

        kwargs_data = sim_util.data_configure_simple(
            numPix=10, deltaPix=1, exposure_time=1, background_rms=1
        )
        data_class = ImageData(**kwargs_data)

        psf_type = "GAUSSIAN"
        fwhm = 0.9
        kwargs_psf = {"psf_type": psf_type, "fwhm": fwhm}
        psf_class = PSF(**kwargs_psf)
        imageModel = ImageLinearFit(
            data_class=data_class,
            psf_class=psf_class,
            lens_model_class=None,
            point_source_class=None,
            source_model_class=sourceModel,
        )

        x_grid, y_grid = util.make_grid(numPix=10, deltapix=1)
        error_map = imageModel.error_map_source(
            kwargs_source=[{"amp": 1}, {"amp": 1}],
            x_grid=x_grid,
            y_grid=y_grid,
            cov_param=np.array([[1, 0], [0, 1]]),
        )
        assert error_map[0] == 2

    def test_create_empty(self):
        kwargs_data = sim_util.data_configure_simple(
            numPix=10, deltaPix=1, exposure_time=1, background_rms=1
        )
        data_class = ImageData(**kwargs_data)
        imageModel_empty = ImageModel(data_class, PSF())
        assert imageModel_empty._psf_error_map == False

        flux = imageModel_empty.lens_surface_brightness(kwargs_lens_light=None)
        assert flux.all() == 0

    def test_extinction_map(self):
        kwargs_data = sim_util.data_configure_simple(
            numPix=10, deltaPix=1, exposure_time=1, background_rms=1
        )
        data_class = ImageData(**kwargs_data)
        extinction_class = DifferentialExtinction(
            optical_depth_model=["UNIFORM"], tau0_index=0
        )
        imageModel = ImageModel(data_class, PSF(), extinction_class=extinction_class)
        extinction = imageModel.extinction_map(
            kwargs_extinction=[{"amp": 1}], kwargs_special={"tau0_list": [1, 0, 0]}
        )
        npt.assert_almost_equal(extinction, np.exp(-1))

    def test_error_response(self):
        C_D_response, model_error = self.imageModel._error_response(
            self.kwargs_lens, self.kwargs_ps, kwargs_special=None
        )
        assert len(model_error) == 100
        print(np.sum(model_error))
        npt.assert_almost_equal(np.sum(model_error), 0.0019271126921470687, decimal=3)

    def test_point_source_linear_response_set(self):
        kwargs_special = {"delta_x_image": [0.1, 0.1], "delta_y_image": [-0.1, -0.1]}
        (
            ra_pos,
            dec_pos,
            amp,
            num_point,
        ) = self.imageModel.point_source_linear_response_set(
            self.kwargs_ps, self.kwargs_lens, kwargs_special, with_amp=True
        )
        ra, dec = self.imageModel.PointSource.image_position(
            self.kwargs_ps, self.kwargs_lens
        )
        npt.assert_almost_equal(ra[0][0], ra_pos[0][0] - 0.1, decimal=5)

    def test_displace_astrometry(self):
        kwargs_special = {
            "delta_x_image": np.array([0.1, 0.1]),
            "delta_y_image": np.array([-0.1, -0.1]),
        }
        x_pos, y_pos = np.array([0, 0]), np.array([0, 0])
        x_shift, y_shift = self.imageModel._displace_astrometry(
            x_pos, y_pos, kwargs_special=kwargs_special
        )
        assert x_pos[0] == 0
        assert x_shift[0] == kwargs_special["delta_x_image"][0]
        assert y_pos[0] == 0
        assert y_shift[0] == kwargs_special["delta_y_image"][0]


if __name__ == "__main__":
    pytest.main()



__author__ = "ajshajib"

from lenstronomy.LensModel.Profiles.gaussian_ellipse_kappa import GaussianEllipseKappa
from lenstronomy.LensModel.Profiles.gaussian_kappa import GaussianKappa

import numpy as np
import numpy.testing as npt
from scipy.special import expi
from scipy.special import wofz
import pytest


class TestGaussianEllipseKappa(object):
    """This class tests the methods for elliptical Gaussian convergence."""

    def setup_method(self):
        """
        :return:
        :rtype:
        """
        self.gaussian_kappa = GaussianKappa()
        self.gaussian_kappa_ellipse = GaussianEllipseKappa()

    def test_function(self):
        """Test the `function()` method at the spherical limit.

        :return:
        :rtype:
        """
        # almost spherical case
        x = 1.0
        y = 1.0
        e1, e2 = 5e-5, 0.0
        sigma = 1.0
        amp = 2.0

        f_ = self.gaussian_kappa_ellipse.function(x, y, amp, sigma, e1, e2)

        r2 = x * x + y * y
        f_sphere = (
            amp
            / (2.0 * np.pi * sigma**2)
            * sigma**2
            * (
                np.euler_gamma
                - expi(-r2 / 2.0 / sigma**2)
                + np.log(r2 / 2.0 / sigma**2)
            )
        )

        npt.assert_almost_equal(f_, f_sphere, decimal=4)

        # spherical case
        e1, e2 = 0.0, 0.0
        f_ = self.gaussian_kappa_ellipse.function(x, y, amp, sigma, e1, e2)

        npt.assert_almost_equal(f_, f_sphere, decimal=4)

    def test_derivatives(self):
        """Test the `derivatives()` method at the spherical limit.

        :return:
        :rtype:
        """
        # almost spherical case
        x = 1.0
        y = 1.0
        e1, e2 = 5e-5, 0.0
        sigma = 1.0
        amp = 2.0

        f_x, f_y = self.gaussian_kappa_ellipse.derivatives(x, y, amp, sigma, e1, e2)
        f_x_sphere, f_y_sphere = self.gaussian_kappa.derivatives(
            x, y, amp=amp, sigma=sigma
        )
        npt.assert_almost_equal(f_x, f_x_sphere, decimal=4)
        npt.assert_almost_equal(f_y, f_y_sphere, decimal=4)

        # spherical case
        e1, e2 = 0.0, 0.0
        f_x, f_y = self.gaussian_kappa_ellipse.derivatives(x, y, amp, sigma, e1, e2)

        npt.assert_almost_equal(f_x, f_x_sphere, decimal=4)
        npt.assert_almost_equal(f_y, f_y_sphere, decimal=4)

    def test_hessian(self):
        """Test the `hessian()` method at the spherical limit.

        :return:
        :rtype:
        """
        # almost spherical case
        x = 1.0
        y = 1.0
        e1, e2 = 5e-5, 0.0
        sigma = 1.0
        amp = 2.0

        f_xx, f_xy, f_yx, f_yy = self.gaussian_kappa_ellipse.hessian(
            x, y, amp, sigma, e1, e2
        )
        (
            f_xx_sphere,
            f_xy_sphere,
            f_yx_sphere,
            f_yy_sphere,
        ) = self.gaussian_kappa.hessian(x, y, amp=amp, sigma=sigma)
        npt.assert_almost_equal(f_xx, f_xx_sphere, decimal=4)
        npt.assert_almost_equal(f_yy, f_yy_sphere, decimal=4)
        npt.assert_almost_equal(f_xy, f_xy_sphere, decimal=4)
        npt.assert_almost_equal(f_yx, f_xy, decimal=8)

        # spherical case
        e1, e2 = 0.0, 0.0
        f_xx, f_xy, f_yx, f_yy = self.gaussian_kappa_ellipse.hessian(
            x, y, amp, sigma, e1, e2
        )

        npt.assert_almost_equal(f_xx, f_xx_sphere, decimal=4)
        npt.assert_almost_equal(f_yy, f_yy_sphere, decimal=4)
        npt.assert_almost_equal(f_xy, f_xy_sphere, decimal=4)

    def test_density_2d(self):
        """Test the `density_2d()` method at the spherical limit.

        :return:
        :rtype:
        """
        # almost spherical case
        x = 1.0
        y = 1.0
        e1, e2 = 5e-5, 0.0
        sigma = 1.0
        amp = 2.0
        f_ = self.gaussian_kappa_ellipse.density_2d(x, y, amp, sigma, e1, e2)
        f_sphere = (
            amp
            / (2.0 * np.pi * sigma**2)
            * np.exp(-(x * x + y * y) / 2.0 / sigma**2)
        )
        npt.assert_almost_equal(f_, f_sphere, decimal=4)

    def test_w_f_approx(self):
        """Test the `w_f_approx()` method with values computed using
        `scipy.special.wofz()`.

        :return:
        :rtype:
        """
        x = np.logspace(-3.0, 3.0, 100)
        y = np.logspace(-3.0, 3.0, 100)

        X, Y = np.meshgrid(x, y)

        w_f_app = self.gaussian_kappa_ellipse.w_f_approx(X + 1j * Y)
        w_f_scipy = wofz(X + 1j * Y)

        npt.assert_allclose(w_f_app.real, w_f_scipy.real, rtol=4e-5, atol=0)
        npt.assert_allclose(w_f_app.imag, w_f_scipy.imag, rtol=4e-5, atol=0)

        # check `derivatives()` method with and without `scipy.special.wofz()`
        x = 1.0
        y = 1.0
        e1, e2 = 5e-5, 0
        sigma = 1.0
        amp = 2.0

        # with `scipy.special.wofz()`
        gauss_scipy = GaussianEllipseKappa(use_scipy_wofz=True)
        f_x_sp, f_y_sp = gauss_scipy.derivatives(x, y, amp, sigma, e1, e2)

        # with `GaussEllipseKappa.w_f_approx()`
        gauss_approx = GaussianEllipseKappa(use_scipy_wofz=False)
        f_x_ap, f_y_ap = gauss_approx.derivatives(x, y, amp, sigma, e1, e2)

        npt.assert_almost_equal(f_x_sp, f_x_ap, decimal=4)
        npt.assert_almost_equal(f_y_sp, f_y_ap, decimal=4)


if __name__ == "__main__":
    pytest.main()



__author__ = "sibirrer"

# this file contains a class to convert lensing and physical units

import numpy as np
import lenstronomy.Util.constants as const
from lenstronomy.Cosmo.background import Background
from lenstronomy.Cosmo.nfw_param import NFWParam

__all__ = ["LensCosmo"]


class LensCosmo(object):
    """Class to manage the physical units and distances present in a single plane lens
    with fixed input cosmology."""

    def __init__(self, z_lens, z_source, cosmo=None):
        """

        :param z_lens: redshift of lens
        :param z_source: redshift of source
        :param cosmo: astropy.cosmology instance
        """

        self.z_lens = z_lens
        self.z_source = z_source
        self.background = Background(cosmo=cosmo)
        self.nfw_param = NFWParam(cosmo=cosmo)

    @property
    def h(self):
        return self.background.cosmo.H(0).value / 100.0

    @property
    def dd(self):
        """

        :return: angular diameter distance to the deflector [Mpc]
        """
        return self.background.d_xy(0, self.z_lens)

    @property
    def ds(self):
        """

        :return: angular diameter distance to the source [Mpc]
        """
        return self.background.d_xy(0, self.z_source)

    @property
    def dds(self):
        """

        :return: angular diameter distance from deflector to source [Mpc]
        """
        return self.background.d_xy(self.z_lens, self.z_source)

    @property
    def ddt(self):
        """

        :return: time delay distance [Mpc]
        """
        return (1 + self.z_lens) * self.dd * self.ds / self.dds

    @property
    def sigma_crit(self):
        """Returns the critical projected lensing mass density in units of M_sun/Mpc^2.

        :return: critical projected lensing mass density
        """
        if not hasattr(self, "_sigma_crit_mpc"):
            const_SI = const.c**2 / (
                4 * np.pi * const.G
            )  # c^2/(4*pi*G) in units of [kg/m]
            conversion = const.Mpc / const.M_sun  # converts [kg/m] to [M_sun/Mpc]
            factor = const_SI * conversion  # c^2/(4*pi*G) in units of [M_sun/Mpc]
            self._sigma_crit_mpc = (
                self.ds / (self.dd * self.dds) * factor
            )  # [M_sun/Mpc^2]
        return self._sigma_crit_mpc

    @property
    def sigma_crit_angle(self):
        """Returns the critical surface density in units of M_sun/arcsec^2 (in physical
        solar mass units) when provided a physical mass per physical Mpc^2.

        :return: critical projected mass density
        """
        if not hasattr(self, "_sigma_crit_arcsec"):
            const_SI = const.c**2 / (
                4 * np.pi * const.G
            )  # c^2/(4*pi*G) in units of [kg/m]
            conversion = const.Mpc / const.M_sun  # converts [kg/m] to [M_sun/Mpc]
            factor = const_SI * conversion  # c^2/(4*pi*G) in units of [M_sun/Mpc]
            self._sigma_crit_arcsec = (
                self.ds / (self.dd * self.dds) * factor * (self.dd * const.arcsec) ** 2
            )  # [M_sun/arcsec^2]
        return self._sigma_crit_arcsec

    def phys2arcsec_lens(self, phys):
        """Convert physical Mpc into arc seconds.

        :param phys: physical distance [Mpc]
        :return: angular diameter [arcsec]
        """
        return phys / self.dd / const.arcsec

    def arcsec2phys_lens(self, arcsec):
        """Convert angular to physical quantities for lens plane.

        :param arcsec: angular size at lens plane [arcsec]
        :return: physical size at lens plane [Mpc]
        """
        return arcsec * const.arcsec * self.dd

    def arcsec2phys_source(self, arcsec):
        """Convert angular to physical quantities for source plane.

        :param arcsec: angular size at source plane [arcsec]
        :return: physical size at source plane [Mpc]
        """
        return arcsec * const.arcsec * self.ds

    def kappa2proj_mass(self, kappa):
        """Convert convergence to projected mass M_sun/Mpc^2.

        :param kappa: lensing convergence
        :return: projected mass [M_sun/Mpc^2]
        """
        return kappa * self.sigma_crit

    def mass_in_theta_E(self, theta_E):
        """Mass within Einstein radius (area * epsilon crit) [M_sun]

        :param theta_E: Einstein radius [arcsec]
        :return: mass within Einstein radius [M_sun]
        """
        mass = self.arcsec2phys_lens(theta_E) ** 2 * np.pi * self.sigma_crit
        return mass

    def mass_in_coin(self, theta_E):
        """

        :param theta_E: Einstein radius [arcsec]
        :return: mass in coin calculated in mean density of the universe
        """
        chi_L = self.background.T_xy(0, self.z_lens)
        chi_S = self.background.T_xy(0, self.z_source)
        return (
            1.0
            / 3
            * np.pi
            * (chi_L * theta_E * const.arcsec) ** 2
            * chi_S
            * self.background.rho_crit
        )  # [M_sun/Mpc**3]

    def time_delay_units(self, fermat_pot, kappa_ext=0):
        """

        :param fermat_pot: in units of arcsec^2 (e.g. Fermat potential)
        :param kappa_ext: unit-less external shear not accounted for in the Fermat potential
        :return: time delay in days
        """
        D_dt = self.ddt * (1.0 - kappa_ext) * const.Mpc  # eqn 7 in Suyu et al.
        return (
            D_dt / const.c * fermat_pot / const.day_s * const.arcsec**2
        )  # * self.arcsec2phys_lens(1.)**2

    def time_delay2fermat_pot(self, dt):
        """

        :param dt: time delay in units of days
        :return: Fermat potential in units arcsec**2 for a given cosmology
        """
        D_dt = self.ddt * const.Mpc
        return dt * const.c * const.day_s / D_dt / const.arcsec**2

    def nfw_angle2physical(self, Rs_angle, alpha_Rs):
        """Converts the angular parameters into the physical ones for an NFW profile.

        :param alpha_Rs: observed bending angle at the scale radius in units of arcsec
        :param Rs_angle: scale radius in units of arcsec
        :return: rho0 [Msun/Mpc^3], Rs [Mpc], c, r200 [Mpc], M200 [Msun]
        """
        Rs = Rs_angle * const.arcsec * self.dd
        theta_scaled = alpha_Rs * self.sigma_crit * self.dd * const.arcsec
        rho0 = theta_scaled / (4 * Rs**2 * (1 + np.log(1.0 / 2.0)))
        rho0_com = rho0 / self.h**2
        c = self.nfw_param.c_rho0(rho0_com, self.z_lens)
        r200 = c * Rs
        M200 = self.nfw_param.M_r200(r200 * self.h, self.z_lens) / self.h
        return rho0, Rs, c, r200, M200

    def nfw_physical2angle(self, M, c):
        """Converts the physical mass and concentration parameter of an NFW profile into
        the lensing quantities.

        :param M: mass enclosed 200 rho_crit in units of M_sun (physical units, meaning
            no little h)
        :param c: NFW concentration parameter (r200/r_s)
        :return: Rs_angle (angle at scale radius) (in units of arcsec), alpha_Rs
            (observed bending angle at the scale radius
        """
        rho0, Rs, r200 = self.nfwParam_physical(M, c)
        Rs_angle = Rs / self.dd / const.arcsec  # Rs in arcsec
        alpha_Rs = rho0 * (4 * Rs**2 * (1 + np.log(1.0 / 2.0)))
        return Rs_angle, alpha_Rs / self.sigma_crit / self.dd / const.arcsec

    def nfwParam_physical(self, M, c):
        """Returns the NFW parameters in physical units.

        :param M: physical mass in M_sun in definition m200
        :param c: concentration
        :return: rho0 [Msun/Mpc^3], Rs [Mpc], r200 [Mpc]
        """
        r200 = (
            self.nfw_param.r200_M(M * self.h, self.z_lens) / self.h
        )  # physical radius r200
        rho0 = (
            self.nfw_param.rho0_c(c, self.z_lens) * self.h**2
        )  # physical density in M_sun/Mpc**3
        Rs = r200 / c
        return rho0, Rs, r200

    def nfw_M_theta_r200(self, M):
        """Returns r200 radius in angular units of arc seconds on the sky.

        :param M: physical mass in M_sun
        :return: angle (in arc seconds) of the r200 radius
        """
        r200 = (
            self.nfw_param.r200_M(M * self.h, self.z_lens) / self.h
        )  # physical radius r200
        theta_r200 = r200 / self.dd / const.arcsec
        return theta_r200

    def sis_theta_E2sigma_v(self, theta_E):
        """Converts the lensing Einstein radius into a physical velocity dispersion.

        :param theta_E: Einstein radius (in arcsec)
        :return: velocity dispersion in units (km/s)
        """
        v_sigma_c2 = theta_E * const.arcsec / (4 * np.pi) * self.ds / self.dds
        return np.sqrt(v_sigma_c2) * const.c / 1000

    def sis_sigma_v2theta_E(self, v_sigma):
        """Converts the velocity dispersion into an Einstein radius for a SIS profile.

        :param v_sigma: velocity dispersion (km/s)
        :return: theta_E (arcsec)
        """
        theta_E = (
            4
            * np.pi
            * (v_sigma * 1000.0 / const.c) ** 2
            * self.dds
            / self.ds
            / const.arcsec
        )
        return theta_E

    def hernquist_phys2angular(self, mass, rs):
        """Translates physical mass definitions of the Hernquist profile to the angular
        units used in the Hernquist lens profile of lenstronomy.

        'sigma0' is defined such that the deflection at projected RS leads to alpha =
        2./3 * Rs * sigma0

        :param mass: A spherical overdensity mass in M_sun corresponding to the mass
            definition mdef at redshift z
        :param rs: rs in units of physical Mpc
        :return: sigma0, Rs_angle
        """
        rs_angle = rs / self.dd / const.arcsec  # Rs in arcsec
        rhos = mass / (2 * np.pi) / rs**3  # units of M_sun / Mpc^3
        sigma0 = rhos * rs  # units of M_sun / Mpc^2
        sigma0 /= self.sigma_crit
        return sigma0, rs_angle

    def hernquist_angular2phys(self, sigma0, rs_angle):
        """'sigma0' is defined such that the deflection at projected RS leads to alpha =
        2./3 * Rs * sigma0.

        :param sigma0: convergence normalization
        :param rs_angle: rs in angular units [arcseconds]
        :return: mass [M_sun], rs  [Mpc]
        """
        rs = rs_angle * self.dd * const.arcsec  # units of Mpc
        rhos = sigma0 / rs * self.sigma_crit
        m_tot = 2 * np.pi * rhos * rs**3
        return m_tot, rs

    def uldm_angular2phys(self, kappa_0, theta_c):
        """Converts the anguar parameters entering the LensModel Uldm() (Ultra Light
        Dark Matter) class in physical masses, i.e. the total soliton mass and the mass
        of the particle.

        :param kappa_0: central convergence of profile
        :param theta_c: core radius (in arcseconds)
        :return: m_eV_log10, M_sol_log10, the log10 of the masses, m in eV and M in
            M_sun
        """
        D_Lens = self.dd * 10**6  # in parsec
        Sigma_c = self.sigma_crit * 10 ** (-12)  # in M_sun / parsec^2
        r_c = theta_c * const.arcsec * D_Lens
        rho0 = 2048 * np.sqrt(0.091) * kappa_0 * Sigma_c / (429 * np.pi * r_c)
        m_log10 = -22 + 0.5 * np.log10(190 / rho0 * (r_c / 100) ** (-4))
        M_log10 = 9 + np.log10(160 * 1.4 / r_c) - 2 * (m_log10 + 22)
        return m_log10, M_log10

    def uldm_mphys2angular(self, m_log10, M_log10):
        """Converts physical ULDM mass in the ones, in angular units, that enter the
        LensModel Uldm() class.

        :param m_log10: exponent of ULDM mass in eV
        :param M_log10: exponent of soliton mass in M_sun
        :return: kappa_0, theta_c, the central convergence and core radius (in
            arcseconds)
        """
        D_Lens = self.dd * 10**6  # in parsec
        Sigma_c = self.sigma_crit * 10 ** (-12)  # in M_sun/parsec^2
        m22 = 10 ** (m_log10 + 22)
        M9 = 10 ** (M_log10 - 9)
        r_c = 160 * 1.4 * m22 ** (-2) * M9 ** (-1)  # core radius in parsec
        rho0 = (
            190 * m22 ** (-2) * (r_c / 100) ** (-4)
        )  # central density in M_sun/parsec^3
        kappa_0 = 429 * np.pi * rho0 * r_c / (2048 * np.sqrt(0.091) * Sigma_c)
        theta_c = r_c / D_Lens / const.arcsec
        return kappa_0, theta_c

    def sersic_m_star2k_eff(self, m_star, R_sersic, n_sersic):
        """Translates a total stellar mass into 'k_eff', the convergence at 'R_sersic'
        (effective radius or half-light radius) for a Sersic profile.

        :param m_star: total stellar mass in physical Msun
        :param R_sersic: half-light radius in arc seconds
        :param n_sersic: Sersic index
        :return: k_eff
        """
        # compute mass integral
        from lenstronomy.LensModel.Profiles.sersic_utils import SersicUtil

        sersic_util = SersicUtil()
        norm_integral = sersic_util.total_flux(
            amp=1, R_sersic=R_sersic, n_sersic=n_sersic
        )
        # compute total kappa normalization and re
        k_eff = m_star / self.sigma_crit_angle
        # renormalize
        k_eff /= norm_integral
        return k_eff

    def sersic_k_eff2m_star(self, k_eff, R_sersic, n_sersic):
        """Translates convergence at half-light radius to total integrated physical
        stellar mass for a Sersic profile.

        :param k_eff: lensing convergence at half-light radius
        :param R_sersic: half-light radius in arc seconds
        :param n_sersic: Sersic index
        :return: stellar mass in physical Msun
        """
        from lenstronomy.LensModel.Profiles.sersic_utils import SersicUtil

        sersic_util = SersicUtil()
        norm_integral = sersic_util.total_flux(
            amp=1, R_sersic=R_sersic, n_sersic=n_sersic
        )
        m_star = k_eff * self.sigma_crit_angle * norm_integral
        return m_star



import numpy as np
from lenstronomy.LensModel.Profiles.spp import SPP
from lenstronomy.LensModel.Profiles.base_profile import LensProfileBase

__all__ = ["CurvedArcSPP", "center_deflector"]


class CurvedArcSPP(LensProfileBase):
    """Lens model that describes a section of a highly magnified deflector region. The
    parameterization is chosen to describe local observables efficient.

    Observables are:
    - curvature radius (basically bending relative to the center of the profile)
    - radial stretch (plus sign) thickness of arc with parity (more generalized than the power-law slope)
    - tangential stretch (plus sign). Infinity means at critical curve
    - direction of curvature
    - position of arc

    Requirements:
    - Should work with other perturbative models without breaking its meaning (say when adding additional shear terms)
    - Must best reflect the observables in lensing
    - minimal covariances between the parameters, intuitive parameterization.
    """

    param_names = [
        "tangential_stretch",
        "radial_stretch",
        "curvature",
        "direction",
        "center_x",
        "center_y",
    ]
    lower_limit_default = {
        "tangential_stretch": -100,
        "radial_stretch": -5,
        "curvature": 0.000001,
        "direction": -np.pi,
        "center_x": -100,
        "center_y": -100,
    }
    upper_limit_default = {
        "tangential_stretch": 100,
        "radial_stretch": 5,
        "curvature": 100,
        "direction": np.pi,
        "center_x": 100,
        "center_y": 100,
    }

    def __init__(self):
        self._spp = SPP()
        super(CurvedArcSPP, self).__init__()

    @staticmethod
    def stretch2spp(
        tangential_stretch, radial_stretch, curvature, direction, center_x, center_y
    ):
        """

        :param tangential_stretch: float, stretch of intrinsic source in tangential direction
        :param radial_stretch: float, stretch of intrinsic source in radial direction
        :param curvature: 1/curvature radius
        :param direction: float, angle in radian
        :param center_x: center of source in image plane
        :param center_y: center of source in image plane
        :return: parameters in terms of a spherical power-law profile resulting in the same observables
        """
        center_x_spp, center_y_spp = center_deflector(
            curvature, direction, center_x, center_y
        )
        r_curvature = 1.0 / curvature
        gamma = (1.0 / radial_stretch - 1) / (1 - 1.0 / tangential_stretch) + 2
        theta_E = abs(1 - 1.0 / tangential_stretch) ** (1.0 / (gamma - 1)) * r_curvature
        return theta_E, gamma, center_x_spp, center_y_spp

    @staticmethod
    def spp2stretch(theta_E, gamma, center_x_spp, center_y_spp, center_x, center_y):
        """Turn Singular power-law lens model into stretch parameterization at position
        (center_x, center_y) This is the inverse function of stretch2spp()

        :param theta_E: Einstein radius of SPP model
        :param gamma: power-law slope
        :param center_x_spp: center of SPP model
        :param center_y_spp: center of SPP model
        :param center_x: center of curved model definition
        :param center_y: center of curved model definition
        :return: tangential_stretch, radial_stretch, curvature, direction
        """
        r_curvature = np.sqrt(
            (center_x_spp - center_x) ** 2 + (center_y_spp - center_y) ** 2
        )
        direction = np.arctan2(center_y - center_y_spp, center_x - center_x_spp)
        tangential_stretch = 1 / (1 - (theta_E / r_curvature) ** (gamma - 1))
        radial_stretch = 1 / (1 + (gamma - 2) * (theta_E / r_curvature) ** (gamma - 1))
        curvature = 1.0 / r_curvature
        return tangential_stretch, radial_stretch, curvature, direction

    def function(
        self,
        x,
        y,
        tangential_stretch,
        radial_stretch,
        curvature,
        direction,
        center_x,
        center_y,
    ):
        """
        ATTENTION: there may not be a global lensing potential!

        :param x:
        :param y:
        :param tangential_stretch: float, stretch of intrinsic source in tangential direction
        :param radial_stretch: float, stretch of intrinsic source in radial direction
        :param curvature: 1/curvature radius
        :param direction: float, angle in radian
        :param center_x: center of source in image plane
        :param center_y: center of source in image plane
        :return:
        """
        theta_E, gamma, center_x_spp, center_y_spp = self.stretch2spp(
            tangential_stretch, radial_stretch, curvature, direction, center_x, center_y
        )
        f_ = self._spp.function(x, y, theta_E, gamma, center_x_spp, center_y_spp)
        alpha_x, alpha_y = self._spp.derivatives(
            center_x, center_y, theta_E, gamma, center_x_spp, center_y_spp
        )
        f_0 = alpha_x * (x - center_x) + alpha_y * (y - center_y)
        return f_ - f_0

    def derivatives(
        self,
        x,
        y,
        tangential_stretch,
        radial_stretch,
        curvature,
        direction,
        center_x,
        center_y,
    ):
        """

        :param x:
        :param y:
        :param tangential_stretch: float, stretch of intrinsic source in tangential direction
        :param radial_stretch: float, stretch of intrinsic source in radial direction
        :param curvature: 1/curvature radius
        :param direction: float, angle in radian
        :param center_x: center of source in image plane
        :param center_y: center of source in image plane
        :return:
        """
        theta_E, gamma, center_x_spp, center_y_spp = self.stretch2spp(
            tangential_stretch, radial_stretch, curvature, direction, center_x, center_y
        )
        f_x, f_y = self._spp.derivatives(
            x, y, theta_E, gamma, center_x_spp, center_y_spp
        )
        f_x0, f_y0 = self._spp.derivatives(
            center_x, center_y, theta_E, gamma, center_x_spp, center_y_spp
        )
        return f_x - f_x0, f_y - f_y0

    def hessian(
        self,
        x,
        y,
        tangential_stretch,
        radial_stretch,
        curvature,
        direction,
        center_x,
        center_y,
    ):
        """

        :param x:
        :param y:
        :param tangential_stretch: float, stretch of intrinsic source in tangential direction
        :param radial_stretch: float, stretch of intrinsic source in radial direction
        :param curvature: 1/curvature radius
        :param direction: float, angle in radian
        :param center_x: center of source in image plane
        :param center_y: center of source in image plane
        :return:
        """
        theta_E, gamma, center_x_spp, center_y_spp = self.stretch2spp(
            tangential_stretch, radial_stretch, curvature, direction, center_x, center_y
        )
        return self._spp.hessian(x, y, theta_E, gamma, center_x_spp, center_y_spp)


def center_deflector(curvature, direction, center_x, center_y):
    """

    :param curvature: 1/curvature radius
    :param direction: float, angle in radian
    :param center_x: center of source in image plane
    :param center_y: center of source in image plane
    :return: center_spp_x, center_spp_y
    """
    center_x_spp = center_x - np.cos(direction) / curvature
    center_y_spp = center_y - np.sin(direction) / curvature
    return center_x_spp, center_y_spp



__author__ = "ajshajib", "sibirrer"
"""
Multi-Gaussian expansion fitting, based on Capellari 2002, http://adsabs.harvard.edu/abs/2002MNRAS.333..400C
"""

import numpy as np
from scipy.optimize import nnls
import warnings
from lenstronomy.Util.package_util import exporter
from lenstronomy.LightModel.Profiles.gaussian import Gaussian

gaussian_func = Gaussian()

export, __all__ = exporter()


@export
def gaussian(R, sigma, amp):
    """

    :param R: radius
    :param sigma: gaussian sigma
    :param amp: normalization
    :return: Gaussian function
    """
    c = amp / (2 * np.pi * sigma**2)
    return c * np.exp(-((R / float(sigma)) ** 2) / 2.0)


@export
def mge_1d(r_array, flux_r, N=20, linspace=False):
    """

    :param r_array: list or radii (numpy array)
    :param flux_r: list of flux values (numpy array)
    :param N: number of Gaussians
    :return: amplitudes and Gaussian sigmas for the best 1d flux profile
    """
    if N == 0:
        warnings.warn(
            "Number of MGE went down to zero! This should not happen!", Warning
        )
        amplitudes = [0]
        sigmas = [1]
        norm = 0
        return amplitudes, sigmas, norm
    try:
        amplitudes, sigmas, norm = _mge_1d(r_array, flux_r, N, linspace=linspace)
    except:
        N_new = N - 1
        amplitudes, sigmas, norm = mge_1d(r_array, flux_r, N=N_new, linspace=linspace)
    return amplitudes, sigmas, norm


@export
def _mge_1d(r_array, flux_r, N=20, linspace=False):
    """

    :param r_array:
    :param flux_r:
    :param N:
    :return:
    """
    if linspace is True:
        sigmas = np.linspace(r_array[0], r_array[-1] / 2, N + 2)[1:-1]
    else:
        sigmas = np.logspace(
            np.log10(r_array[0]), np.log10((r_array[-1] + 0.0000001) / 2.0), N + 2
        )[1:-1]
    # sigmas = np.linspace(r_array[0], r_array[-1]/2, N + 2)[1:-1]

    A = np.zeros((len(flux_r), N))
    for j in np.arange(A.shape[1]):
        A[:, j] = gaussian(r_array, sigmas[j], 1.0)
    amplitudes, norm = nnls(A, flux_r)
    return amplitudes, sigmas, norm


@export
def de_projection_3d(amplitudes, sigmas):
    """De-projects a gaussian (or list of multiple Gaussians from a 2d projected to a 3d
    profile) :param amplitudes:

    :param sigmas:
    :return:
    """
    amplitudes_3d = amplitudes / sigmas / np.sqrt(2 * np.pi)
    return amplitudes_3d, sigmas



from lenstronomy.Workflow.psf_fitting import PsfFitting
from lenstronomy.Workflow.alignment_matching import AlignmentFitting
from lenstronomy.Workflow.flux_calibration import FluxCalibration
from lenstronomy.ImSim.MultiBand.single_band_multi_model import SingleBandMultiModel
from lenstronomy.Workflow.multi_band_manager import MultiBandUpdateManager
from lenstronomy.Sampling.likelihood import LikelihoodModule
from lenstronomy.Sampling.sampler import Sampler
from lenstronomy.Sampling.Samplers.multinest_sampler import MultiNestSampler
from lenstronomy.Sampling.Samplers.polychord_sampler import DyPolyChordSampler
from lenstronomy.Sampling.Samplers.dynesty_sampler import DynestySampler
from lenstronomy.Sampling.Samplers.nautilus_sampler import NautilusSampler
from lenstronomy.Sampling.Samplers.cobaya_sampler import CobayaSampler
import numpy as np
import lenstronomy.Util.analysis_util as analysis_util

__all__ = ["FittingSequence"]


class FittingSequence(object):
    """Class to define a sequence of fitting applied, inherit the Fitting class this is
    a Workflow manager that allows to update model configurations before executing
    another step in the modelling The user can take this module as an example of how to
    create their own workflows or build their own around the FittingSequence."""

    def __init__(
        self,
        kwargs_data_joint,
        kwargs_model,
        kwargs_constraints,
        kwargs_likelihood,
        kwargs_params,
        mpi=False,
        verbose=True,
    ):
        """

        :param kwargs_data_joint: keyword argument specifying the data according to LikelihoodModule
        :param kwargs_model: keyword arguments to describe all model components used in
         class_creator.create_class_instances()
        :param kwargs_constraints: keyword arguments of the Param() class to handle parameter constraints during the
         sampling (except upper and lower limits and sampling input mean and width)
        :param kwargs_likelihood: keyword arguments of the Likelihood() class to handle parameters and settings of the
         likelihood
        :param kwargs_params: setting of the sampling bounds and initial guess mean and spread.
         The argument is organized as:
         'lens_model': [kwargs_init, kwargs_sigma, kwargs_fixed, kwargs_lower, kwargs_upper]
         'source_model': [kwargs_init, kwargs_sigma, kwargs_fixed, kwargs_lower, kwargs_upper]
         'lens_light_model': [kwargs_init, kwargs_sigma, kwargs_fixed, kwargs_lower, kwargs_upper]
         'point_source_model': [kwargs_init, kwargs_sigma, kwargs_fixed, kwargs_lower, kwargs_upper]
         'extinction_model': [kwargs_init, kwargs_sigma, kwargs_fixed, kwargs_lower, kwargs_upper]
         'special': [kwargs_init, kwargs_sigma, kwargs_fixed, kwargs_lower, kwargs_upper]
        :param mpi: MPI option (bool), if True, will launch an MPI Pool job for the steps in the fitting sequence where
         possible
        :param verbose: bool, if True prints temporary results and indicators of the fitting process
        """
        self.kwargs_data_joint = kwargs_data_joint
        self.multi_band_list = kwargs_data_joint.get("multi_band_list", [])
        self.multi_band_type = kwargs_data_joint.get("multi_band_type", "single-band")
        self._verbose = verbose
        self._mpi = mpi
        self._updateManager = MultiBandUpdateManager(
            kwargs_model,
            kwargs_constraints,
            kwargs_likelihood,
            kwargs_params,
            num_bands=len(self.multi_band_list),
        )
        self._mcmc_init_samples = None

    @property
    def kwargs_fixed(self):
        """Returns the updated kwargs_fixed from the update manager.

        :return: list of fixed kwargs, see UpdateManager()
        """
        return self._updateManager.fixed_kwargs

    def fit_sequence(self, fitting_list):
        """

        :param fitting_list: list of [['string', {kwargs}], ..] with 'string being the specific fitting option and
         kwargs being the arguments passed to this option
        :return: fitting results
        """
        chain_list = []
        for i, fitting in enumerate(fitting_list):
            fitting_type = fitting[0]
            kwargs = fitting[1]

            if fitting_type == "restart":
                self._updateManager.set_init_state()

            elif fitting_type == "update_settings":
                self.update_settings(**kwargs)

            elif fitting_type == "set_param_value":
                self.set_param_value(**kwargs)

            elif fitting_type == "fix_not_computed":
                self.fix_not_computed(**kwargs)

            elif fitting_type == "psf_iteration":
                self.psf_iteration(**kwargs)

            elif fitting_type == "align_images":
                self.align_images(**kwargs)

            elif fitting_type == "calibrate_images":
                self.flux_calibration(**kwargs)

            elif fitting_type == "PSO":
                kwargs_result, chain, param = self.pso(**kwargs)
                self._updateManager.update_param_state(**kwargs_result)
                chain_list.append([fitting_type, chain, param])

            elif fitting_type == "SIMPLEX":
                kwargs_result = self.simplex(**kwargs)
                self._updateManager.update_param_state(**kwargs_result)
                chain_list.append([fitting_type, kwargs_result])

            elif fitting_type == "MCMC":
                if "init_samples" not in kwargs:
                    kwargs["init_samples"] = self._mcmc_init_samples
                elif kwargs["init_samples"] is None:
                    kwargs["init_samples"] = self._mcmc_init_samples
                mcmc_output = self.mcmc(**kwargs)
                kwargs_result = self._result_from_mcmc(mcmc_output)
                self._updateManager.update_param_state(**kwargs_result)
                chain_list.append(mcmc_output)

            elif fitting_type == "Nautilus":
                # do importance nested sampling with Nautilus
                nautilus = NautilusSampler(
                    likelihood_module=self.likelihoodModule, mpi=self._mpi, **kwargs
                )
                points, log_w, log_l, log_z = nautilus.run(**kwargs)
                chain_list.append([points, log_w, log_l, log_z])
                if kwargs.get("verbose", False):
                    print(len(points), "number of points sampled")
                kwargs_result = self.best_fit_from_samples(points, log_l)
                self._updateManager.update_param_state(**kwargs_result)

            elif fitting_type == "nested_sampling":
                ns_output = self.nested_sampling(**kwargs)
                chain_list.append(ns_output)

            elif fitting_type == "metropolis_hastings":
                print("Using the Metropolis--Hastings MCMC sampler in Cobaya.")

                param_class = self.param_class

                kwargs_temp = self._updateManager.parameter_state
                mean_start = param_class.kwargs2args(**kwargs_temp)
                kwargs_sigma = self._updateManager.sigma_kwargs
                sigma_start = np.array(param_class.kwargs2args(**kwargs_sigma))

                # pass the likelihood and starting info to the sampler
                sampler = CobayaSampler(self.likelihoodModule, mean_start, sigma_start)

                # run the sampler
                updated_info, sampler_type, best_fit_values = sampler.run(**kwargs)

                # change the best-fit values returned by cobaya into lenstronomy kwargs format
                best_fit_kwargs = self.param_class.args2kwargs(
                    best_fit_values, bijective=True
                )

                # collect the products
                mh_output = [updated_info, sampler_type, best_fit_kwargs]

                # append the products to the chain list
                chain_list.append(mh_output)

            else:
                raise ValueError(
                    "fitting_sequence {} is not supported. Please use: 'PSO', 'SIMPLEX', 'MCMC', 'metropolis_hastings', "
                    "'Nautilus', 'nested_sampling', 'psf_iteration', 'restart', 'update_settings', 'calibrate_images' or "
                    "'align_images'".format(fitting_type)
                )
        return chain_list

    def best_fit(self, bijective=False):
        """

        :param bijective: bool, if True, the mapping of image2source_plane and the mass_scaling parameterisation are inverted. If you do not use those options, there is no effect.
        :return: best fit model of the current state of the FittingSequence class
        """

        return self._updateManager.best_fit(bijective=bijective)

    def update_state(self, kwargs_update):
        """Updates current best fit state to the input model keywords specified.

        :param kwargs_update: format of kwargs_result
        :return: None
        """
        self._updateManager.update_param_state(**kwargs_update)

    @property
    def best_fit_likelihood(self):
        """Returns the log likelihood of the best fit model of the current state of this
        class.

        :return: log likelihood, float
        """
        kwargs_result = self.best_fit(bijective=False)
        param_class = self.param_class
        likelihoodModule = self.likelihoodModule
        logL = likelihoodModule.logL(param_class.kwargs2args(**kwargs_result))
        return logL

    @property
    def bic(self):
        """Bayesian information criterion (BIC) of the model.

        :return: bic value, float
        """
        num_data = self.likelihoodModule.num_data
        num_param_nonlinear = self.param_class.num_param()[0]
        num_param_linear = self.param_class.num_param_linear()
        num_param = num_param_nonlinear + num_param_linear
        bic = analysis_util.bic_model(self.best_fit_likelihood, num_data, num_param)
        return bic

    @property
    def param_class(self):
        """

        :return: Param() class instance reflecting the current state of FittingSequence
        """
        return self._updateManager.param_class

    @property
    def likelihoodModule(self):
        """

        :return: Likelihood() class instance reflecting the current state of FittingSequence
        """
        kwargs_model = self._updateManager.kwargs_model
        kwargs_likelihood = self._updateManager.kwargs_likelihood
        likelihoodModule = LikelihoodModule(
            self.kwargs_data_joint, kwargs_model, self.param_class, **kwargs_likelihood
        )
        return likelihoodModule

    def simplex(self, n_iterations, method="Nelder-Mead"):
        """Downhill simplex optimization using the Nelder-Mead algorithm.

        :param n_iterations: maximum number of iterations to perform
        :param method: the optimization method used, see documentation in
            scipy.optimize.minimize
        :return: result of the best fit
        """

        param_class = self.param_class
        kwargs_temp = self._updateManager.parameter_state
        init_pos = param_class.kwargs2args(**kwargs_temp)
        sampler = Sampler(likelihoodModule=self.likelihoodModule)
        result = sampler.simplex(init_pos, n_iterations, method)

        kwargs_result = param_class.args2kwargs(result, bijective=True)
        return kwargs_result

    def mcmc(
        self,
        n_burn,
        n_run,
        walkerRatio=None,
        n_walkers=None,
        sigma_scale=1,
        threadCount=1,
        init_samples=None,
        re_use_samples=True,
        sampler_type="EMCEE",
        progress=True,
        backend_filename=None,
        start_from_backend=False,
        **kwargs_zeus
    ):
        """MCMC routine.

        :param n_burn: number of burn in iterations (will not be saved)
        :param n_run: number of MCMC iterations that are saved
        :param walkerRatio: ratio of walkers/number of free parameters
        :param n_walkers: integer, number of walkers of emcee (optional, if set, overwrites the walkerRatio input
        :param sigma_scale: scaling of the initial parameter spread relative to the width in the initial settings
        :param threadCount: number of CPU threads. If MPI option is set, threadCount=1
        :param init_samples: initial sample from where to start the MCMC process
        :param re_use_samples: bool, if True, re-uses the samples described in init_samples.nOtherwise starts from
         scratch.
        :param sampler_type: string, which MCMC sampler to be used. Options are: 'EMCEE', 'ZEUS'
        :param progress: boolean, if True shows progress bar in EMCEE
        :param backend_filename: name of the HDF5 file where sampling state is saved (through emcee backend engine)
        :type backend_filename: string
        :param start_from_backend: if True, start from the state saved in `backup_filename`.
         O therwise, create a new backup file with name `backup_filename` (any already existing file is overwritten!).
        :type start_from_backend: bool
        :param kwargs_zeus: zeus-specific kwargs
        :return: list of output arguments, e.g. MCMC samples, parameter names, logL distances of all samples specified
         by the specific sampler used
        """

        param_class = self.param_class
        # run PSO
        mcmc_class = Sampler(likelihoodModule=self.likelihoodModule)
        kwargs_temp = self._updateManager.parameter_state
        mean_start = param_class.kwargs2args(**kwargs_temp)
        kwargs_sigma = self._updateManager.sigma_kwargs
        sigma_start = np.array(param_class.kwargs2args(**kwargs_sigma)) * sigma_scale
        num_param, param_list = param_class.num_param()
        if n_walkers is None:
            if walkerRatio is None:
                raise ValueError(
                    "MCMC sampler needs either n_walkers or walkerRatio as input argument"
                )
            n_walkers = num_param * walkerRatio
        # run MCMC
        if init_samples is not None and re_use_samples is True:
            num_samples, num_param_prev = np.shape(init_samples)
            if num_param_prev == num_param:
                print("re-using previous samples to initialize the next MCMC run.")
                idxs = np.random.choice(len(init_samples), n_walkers)
                initpos = init_samples[idxs]
            else:
                raise ValueError(
                    "Can not re-use previous MCMC samples as number of parameters have changed!"
                )
        else:
            initpos = None

        if sampler_type == "EMCEE":
            samples, dist = mcmc_class.mcmc_emcee(
                n_walkers,
                n_run,
                n_burn,
                mean_start,
                sigma_start,
                mpi=self._mpi,
                threadCount=threadCount,
                progress=progress,
                initpos=initpos,
                backend_filename=backend_filename,
                start_from_backend=start_from_backend,
            )
            output = [sampler_type, samples, param_list, dist]

        elif sampler_type == "ZEUS":
            samples, dist = mcmc_class.mcmc_zeus(
                n_walkers,
                n_run,
                n_burn,
                mean_start,
                sigma_start,
                mpi=self._mpi,
                threadCount=threadCount,
                progress=progress,
                initpos=initpos,
                backend_filename=backend_filename,
                **kwargs_zeus
            )
            output = [sampler_type, samples, param_list, dist]
        else:
            raise ValueError("sampler_type %s not supported!" % sampler_type)
        self._mcmc_init_samples = samples  # overwrites previous samples to continue from there in the next MCMC run
        return output

    def pso(
        self, n_particles, n_iterations, sigma_scale=1, print_key="PSO", threadCount=1
    ):
        """Particle Swarm Optimization.

        :param n_particles: number of particles in the Particle Swarm Optimization
        :param n_iterations: number of iterations in the optimization process
        :param sigma_scale: scaling of the initial parameter spread relative to the
            width in the initial settings
        :param print_key: string, printed text when executing this routine
        :param threadCount: number of CPU threads. If MPI option is set, threadCount=1
        :return: result of the best fit, the PSO chain of the best fit parameter after
            each iteration [lnlikelihood, parameters, velocities], list of parameters in
            same order as in chain
        """

        param_class = self.param_class
        kwargs_temp = self._updateManager.parameter_state
        init_pos = param_class.kwargs2args(**kwargs_temp)
        kwargs_sigma = self._updateManager.sigma_kwargs
        sigma_start = param_class.kwargs2args(**kwargs_sigma)
        lower_start = np.array(init_pos) - np.array(sigma_start) * sigma_scale
        upper_start = np.array(init_pos) + np.array(sigma_start) * sigma_scale

        num_param, param_list = param_class.num_param()
        # run PSO
        sampler = Sampler(likelihoodModule=self.likelihoodModule)
        result, chain = sampler.pso(
            n_particles,
            n_iterations,
            lower_start,
            upper_start,
            init_pos=init_pos,
            threadCount=threadCount,
            mpi=self._mpi,
            print_key=print_key,
        )
        kwargs_result = param_class.args2kwargs(result, bijective=True)
        return kwargs_result, chain, param_list

    def nested_sampling(
        self,
        sampler_type="MULTINEST",
        kwargs_run={},
        prior_type="uniform",
        width_scale=1,
        sigma_scale=1,
        output_basename="chain",
        remove_output_dir=True,
        dypolychord_dynamic_goal=0.8,
        polychord_settings={},
        dypolychord_seed_increment=200,
        output_dir="nested_sampling_chains",
        dynesty_bound="multi",
        dynesty_sample="auto",
    ):
        """Run (Dynamic) Nested Sampling algorithms, depending on the type of algorithm.

        :param sampler_type: 'MULTINEST', 'DYPOLYCHORD', 'DYNESTY'
        :param kwargs_run: keywords passed to the core sampling method
        :param prior_type: 'uniform' of
            'gaussian', for converting the unit hypercube to param cube :param width_scale:
            scale the width (lower/upper limits) of the parameters space by this factor
        :param sigma_scale: if prior_type is 'gaussian', scale the gaussian sigma by
            this factor
        :param output_basename: name of the folder in which the core
            MultiNest/PolyChord code will save output files
        :param remove_output_dir: if True, the above folder is removed after completion
        :param dypolychord_dynamic_goal: dynamic goal for DyPolyChord (trade-off between
            evidence (0) and posterior (1) computation) :param polychord_settings: settings
            dictionary to send to pypolychord. Check dypolychord documentation for details.
        :param dypolychord_seed_increment: seed increment for dypolychord with MPI.
            Check dypolychord documentation for details.
        :param dynesty_bound: see https://dynesty.readthedocs.io
        :param sampler_type: 'MULTINEST', 'DYPOLYCHORD', 'DYNESTY'
        :param kwargs_run: keywords passed to the core sampling method
        :param prior_type: 'uniform' of 'gaussian', for converting the unit hypercube to
            param cube
        :param width_scale: scale the width (lower/upper limits) of the parameters space
            by this factor
        :param sigma_scale: if prior_type is 'gaussian', scale the gaussian sigma by
            this factor
        :param output_basename: name of the folder in which the core MultiNest/PolyChord
            code will save output files
        :param remove_output_dir: if True, the above folder is removed after completion
        :param dypolychord_dynamic_goal: dynamic goal for DyPolyChord (trade-off between
            evidence (0) and posterior (1) computation)
        :param polychord_settings: settings dictionary to send to pypolychord. Check
            dypolychord documentation for details.
        :param dypolychord_seed_increment: seed increment for dypolychord with MPI.
            Check dypolychord documentation for details.
        :param dynesty_bound: see https://dynesty.readthedocs.io for details
        :param dynesty_sample: see https://dynesty.readthedocs.io for details
        :return: list of output arguments : samples, mean inferred values, log-
            likelihood, log-evidence, error on log-evidence for each sample
        """
        mean_start, sigma_start = self._prepare_sampling(prior_type)

        if sampler_type == "MULTINEST":
            sampler = MultiNestSampler(
                self.likelihoodModule,
                prior_type=prior_type,
                prior_means=mean_start,
                prior_sigmas=sigma_start,
                width_scale=width_scale,
                sigma_scale=sigma_scale,
                output_dir=output_dir,
                output_basename=output_basename,
                remove_output_dir=remove_output_dir,
                use_mpi=self._mpi,
            )
            samples, means, logZ, logZ_err, logL, results_object = sampler.run(
                kwargs_run
            )

        elif sampler_type == "DYPOLYCHORD":
            if "resume_dyn_run" in kwargs_run and kwargs_run["resume_dyn_run"] is True:
                resume_dyn_run = True
            else:
                resume_dyn_run = False
            sampler = DyPolyChordSampler(
                self.likelihoodModule,
                prior_type=prior_type,
                prior_means=mean_start,
                prior_sigmas=sigma_start,
                width_scale=width_scale,
                sigma_scale=sigma_scale,
                output_dir=output_dir,
                output_basename=output_basename,
                polychord_settings=polychord_settings,
                remove_output_dir=remove_output_dir,
                resume_dyn_run=resume_dyn_run,
                use_mpi=self._mpi,
            )
            samples, means, logZ, logZ_err, logL, results_object = sampler.run(
                dypolychord_dynamic_goal, kwargs_run
            )

        elif sampler_type == "DYNESTY":
            sampler = DynestySampler(
                self.likelihoodModule,
                prior_type=prior_type,
                prior_means=mean_start,
                prior_sigmas=sigma_start,
                width_scale=width_scale,
                sigma_scale=sigma_scale,
                bound=dynesty_bound,
                sample=dynesty_sample,
                use_mpi=self._mpi,
            )
            samples, means, logZ, logZ_err, logL, results_object = sampler.run(
                kwargs_run
            )

        else:
            raise ValueError("Sampler type %s not supported." % sampler_type)
        # update current best fit values
        self._update_state(samples[-1])

        output = [
            sampler_type,
            samples,
            sampler.param_names,
            logL,
            logZ,
            logZ_err,
            results_object,
        ]
        return output

    def psf_iteration(self, compute_bands=None, **kwargs_psf_iter):
        """Iterative PSF reconstruction.

        :param compute_bands: bool list, if multiple bands, this process can be limited
            to a subset of bands
        :param kwargs_psf_iter: keyword arguments as used or available in
            PSFIteration.update_iterative() definition
        :return: 0, updated PSF is stored in self.multi_band_list
        """
        kwargs_model = self._updateManager.kwargs_model
        kwargs_likelihood = self._updateManager.kwargs_likelihood
        likelihood_mask_list = kwargs_likelihood.get("image_likelihood_mask_list", None)
        kwargs_pixelbased = kwargs_likelihood.get("kwargs_pixelbased", None)
        kwargs_temp = self.best_fit(bijective=False)
        if compute_bands is None:
            compute_bands = [True] * len(self.multi_band_list)

        for band_index in range(len(self.multi_band_list)):
            if compute_bands[band_index] is True:
                kwargs_psf = self.multi_band_list[band_index][1]
                image_model = SingleBandMultiModel(
                    self.multi_band_list,
                    kwargs_model,
                    likelihood_mask_list=likelihood_mask_list,
                    band_index=band_index,
                    kwargs_pixelbased=kwargs_pixelbased,
                )
                psf_iter = PsfFitting(image_model_class=image_model)
                kwargs_psf = psf_iter.update_iterative(
                    kwargs_psf, kwargs_params=kwargs_temp, **kwargs_psf_iter
                )
                self.multi_band_list[band_index][1] = kwargs_psf
        return 0

    def align_images(
        self,
        n_particles=10,
        n_iterations=10,
        align_offset=True,
        align_rotation=False,
        threadCount=1,
        compute_bands=None,
        delta_shift=0.2,
        delta_rot=0.1,
    ):
        """Aligns the coordinate systems of different exposures within a fixed model
        parameterisation by executing a PSO with relative coordinate shifts as free
        parameters.

        :param n_particles: number of particles in the Particle Swarm Optimization
        :param n_iterations: number of iterations in the optimization process
        :param align_offset: aligns shift in Ra and Dec
        :type align_offset: boolean
        :param align_rotation: aligns coordinate rotation
        :type align_rotation: boolean
        :param delta_shift: astrometric shift tolerance
        :param delta_rot: rotation angle tolerance [in radian]
        :param compute_bands: bool list, if multiple bands, this process can be limited
            to a subset of bands for which the coordinate system is being fit for best
            alignment to the model parameters
        :return: 0, updated coordinate system for the band(s)
        """
        kwargs_model = self._updateManager.kwargs_model
        kwargs_likelihood = self._updateManager.kwargs_likelihood
        likelihood_mask_list = kwargs_likelihood.get("image_likelihood_mask_list", None)
        kwargs_temp = self.best_fit(bijective=False)
        if compute_bands is None:
            compute_bands = [True] * len(self.multi_band_list)

        for i in range(len(self.multi_band_list)):
            if compute_bands[i] is True:
                alignmentFitting = AlignmentFitting(
                    self.multi_band_list,
                    kwargs_model,
                    kwargs_temp,
                    band_index=i,
                    likelihood_mask_list=likelihood_mask_list,
                    align_offset=align_offset,
                    align_rotation=align_rotation,
                )

                kwargs_data, chain = alignmentFitting.pso(
                    n_particles=n_particles,
                    n_iterations=n_iterations,
                    delta_shift=delta_shift,
                    delta_rot=delta_rot,
                    threadCount=threadCount,
                    mpi=self._mpi,
                    print_key="Alignment fitting for band %s ..." % i,
                )
                print("Align completed for band %s." % i)
                print(
                    "ra_shift: %s,  dec_shift: %s, phi_rot: %s"
                    % (
                        kwargs_data.get("ra_shift", 0),
                        kwargs_data.get("dec_shift", 0),
                        kwargs_data.get("phi_rot", 0),
                    )
                )
                self.multi_band_list[i][0] = kwargs_data
        return 0

    def flux_calibration(
        self,
        n_particles=10,
        n_iterations=10,
        threadCount=1,
        calibrate_bands=None,
        scaling_lower_limit=0,
        scaling_upper_limit=1000,
    ):
        """Calibrates flux_scaling between multiple images. This routine only works in
        'join-linear' model when fluxes are meant to be identical for different bands.

        :param n_particles: number of particles in the Particle Swarm Optimization
        :param n_iterations: number of iterations in the optimization process
        :param calibrate_bands: state which bands the flux calibration is applied to
        :type calibrate_bands: list of booleans of length of the imaging bands
        :param threadCount: number of CPU threads. If MPI option is set, threadCount=1
        :type threadCount: integer
        :param scaling_lower_limit: lower limit of flux_scaling
        :param scaling_upper_limit: upper limit of flux scaling
        :return: 0, updated coordinate system for the band(s)
        """
        kwargs_model = self._updateManager.kwargs_model
        kwargs_temp = self.best_fit(bijective=False)
        multi_band_type = self.kwargs_data_joint.get("multi_band_type", "multi-linear")
        kwargs_imaging = self.likelihoodModule.kwargs_imaging

        calibration_fitting = FluxCalibration(
            kwargs_imaging=kwargs_imaging,
            kwargs_model=kwargs_model,
            kwargs_params=kwargs_temp,
            calibrate_bands=calibrate_bands,
        )

        multi_band_list, chain = calibration_fitting.pso(
            n_particles=n_particles,
            n_iterations=n_iterations,
            threadCount=threadCount,
            mpi=self._mpi,
            scaling_lower_limit=scaling_lower_limit,
            scaling_upper_limit=scaling_upper_limit,
        )
        self.multi_band_list = multi_band_list
        return 0

    def update_settings(
        self,
        kwargs_model=None,
        kwargs_constraints=None,
        kwargs_likelihood=None,
        lens_add_fixed=None,
        source_add_fixed=None,
        lens_light_add_fixed=None,
        ps_add_fixed=None,
        special_add_fixed=None,
        lens_remove_fixed=None,
        source_remove_fixed=None,
        lens_light_remove_fixed=None,
        ps_remove_fixed=None,
        special_remove_fixed=None,
        change_source_lower_limit=None,
        change_source_upper_limit=None,
        change_lens_lower_limit=None,
        change_lens_upper_limit=None,
        change_sigma_lens=None,
        change_sigma_source=None,
        change_sigma_lens_light=None,
    ):
        """Updates lenstronomy settings "on the fly".

        :param kwargs_model: kwargs, specified keyword arguments overwrite the existing ones
        :param kwargs_constraints: kwargs, specified keyword arguments overwrite the existing ones
        :param kwargs_likelihood: kwargs, specified keyword arguments overwrite the existing ones
        :param lens_add_fixed: [[i_model, ['param1', 'param2',...], [...]]
        :param source_add_fixed: [[i_model, ['param1', 'param2',...], [...]]
        :param lens_light_add_fixed: [[i_model, ['param1', 'param2',...], [...]]
        :param ps_add_fixed: [[i_model, ['param1', 'param2',...], [...]]
        :param special_add_fixed: ['param1', 'param2',...]
        :param lens_remove_fixed: [[i_model, ['param1', 'param2',...], [...]]
        :param source_remove_fixed: [[i_model, ['param1', 'param2',...], [...]]
        :param lens_light_remove_fixed: [[i_model, ['param1', 'param2',...], [...]]
        :param ps_remove_fixed: [[i_model, ['param1', 'param2',...], [...]]
        :param special_remove_fixed: ['param1', 'param2',...]
        :param change_lens_lower_limit: [[i_model, ['param_name1', 'param_name2', ...], [value1, value2, ...]]]
        :param change_lens_upper_limit: [[i_model, ['param_name1', 'param_name2', ...], [value1, value2, ...]]]
        :param change_source_lower_limit: [[i_model, ['param_name1', 'param_name2', ...], [value1, value2, ...]]]
        :param change_source_upper_limit: [[i_model, [''param_name1', 'param_name2', ...], [value1, value2, ...]]]
        :param change_sigma_lens: [[i_model, ['param_name1', 'param_name2', ...], [value1, value2, ...]]]
        :param change_sigma_source: [[i_model, ['param_name1', 'param_name2', ...], [value1, value2, ...]]]
        :param change_sigma_lens_light: [[i_model, ['param_name1', 'param_name2', ...], [value1, value2, ...]]]
        :return: 0, the settings are overwritten for the next fitting step to come
        """
        self._updateManager.update_options(
            kwargs_model, kwargs_constraints, kwargs_likelihood
        )
        self._updateManager.update_fixed(
            lens_add_fixed,
            source_add_fixed,
            lens_light_add_fixed,
            ps_add_fixed,
            special_add_fixed,
            lens_remove_fixed,
            source_remove_fixed,
            lens_light_remove_fixed,
            ps_remove_fixed,
            special_remove_fixed,
        )
        self._updateManager.update_limits(
            change_source_lower_limit,
            change_source_upper_limit,
            change_lens_lower_limit,
            change_lens_upper_limit,
        )
        self._updateManager.update_sigmas(
            change_sigma_lens=change_sigma_lens,
            change_sigma_source=change_sigma_source,
            change_sigma_lens_light=change_sigma_lens_light,
        )
        return 0

    def set_param_value(self, **kwargs):
        """Set a parameter to a specific value. `kwargs` are below.

        :param lens: [[i_model, ['param1', 'param2',...], [...]]
        :type lens:
        :param source: [[i_model, ['param1', 'param2',...], [...]]
        :type source:
        :param lens_light: [[i_model, ['param1', 'param2',...], [...]]
        :type lens_light:
        :param ps: [[i_model, ['param1', 'param2',...], [...]]
        :type ps:
        :return: 0, the value of the param is overwritten
        :rtype:
        """
        self._updateManager.update_param_value(**kwargs)

    def fix_not_computed(self, free_bands):
        """Fixes lens model parameters of imaging bands/frames that are not computed and
        frees the parameters of the other lens models to the initial kwargs_fixed
        options.

        :param free_bands: bool list of length of imaging bands in order of imaging
            bands, if False: set fixed lens model
        :return: None
        """
        self._updateManager.fix_not_computed(free_bands=free_bands)

    def _prepare_sampling(self, prior_type):
        if prior_type == "gaussian":
            mean_start = self.param_class.kwargs2args(
                **self._updateManager.parameter_state
            )
            sigma_start = self.param_class.kwargs2args(
                **self._updateManager.sigma_kwargs
            )
            mean_start = np.array(mean_start)
            sigma_start = np.array(sigma_start)
        else:
            mean_start, sigma_start = None, None
        return mean_start, sigma_start

    def _update_state(self, result):
        """

        :param result: array of parameters being sampled (e.g. result of MCMC chain)
        :return: None, updates the parameter state of the class instance
        """
        kwargs_result = self.param_class.args2kwargs(result, bijective=True)
        self._updateManager.update_param_state(**kwargs_result)

    def _result_from_mcmc(self, mcmc_output):
        """

        :param mcmc_output: list returned by self.mcmc()
        :return: kwargs_result like returned by self.pso(), from best logL MCMC sample
        """
        _, samples, _, logL_values = mcmc_output
        return self.best_fit_from_samples(samples, logL_values)

    def best_fit_from_samples(self, samples, logl):
        """Return best fit (max likelihood) value of samples in lenstronomy conventions.

        :param samples: samples of multi-dimensional parameter space
        :param logl: likelihood values for each sample
        :return: kwargs_result in lenstronomy convention
        """
        # get index of best logL sample
        best_fit_index = np.argmax(logl)
        best_fit_sample = samples[best_fit_index, :]
        best_fit_result = best_fit_sample.tolist()
        # get corresponding kwargs
        kwargs_result = self.param_class.args2kwargs(best_fit_result, bijective=True)
        return kwargs_result



__author__ = "sibirrer"

import pytest
import numpy as np
import numpy.testing as npt

from lenstronomy.LensModel.Profiles.hernquist import Hernquist
from lenstronomy.LensModel.Profiles.hernquist_ellipse import Hernquist_Ellipse
import lenstronomy.Util.param_util as param_util


class TestHernquist(object):
    def setup_method(self):
        self.profile = Hernquist()

    def test_function(self):
        x = np.array([1])
        y = np.array([2])
        Rs = 1.0
        sigma0 = 0.5
        values = self.profile.function(x, y, sigma0, Rs)
        npt.assert_almost_equal(values[0], 0.66514613455415028, decimal=8)
        x = np.array([0])
        y = np.array([0])
        Rs = 1.0
        sigma0 = 0.5
        values = self.profile.function(x, y, sigma0, Rs)
        npt.assert_almost_equal(values[0], 0, decimal=6)

        x = np.array([2, 3, 4])
        y = np.array([1, 1, 1])
        values = self.profile.function(x, y, sigma0, Rs)
        npt.assert_almost_equal(values[0], 0.66514613455415028, decimal=8)
        npt.assert_almost_equal(values[1], 0.87449395673649566, decimal=8)
        npt.assert_almost_equal(values[2], 1.0549139073851708, decimal=8)

    def test_derivatives(self):
        x = 1
        y = 2
        Rs = 1.0
        sigma0 = 0.5
        f_x, f_y = self.profile.derivatives(x, y, sigma0, Rs)
        npt.assert_almost_equal(f_x, 0.11160641027573866, decimal=8)
        npt.assert_almost_equal(f_y, 0.22321282055147731, decimal=8)
        x = np.array([0])
        y = np.array([0])
        f_x, f_y = self.profile.derivatives(x, y, sigma0, Rs)
        npt.assert_almost_equal(f_x, 0, decimal=8)
        npt.assert_almost_equal(f_y, 0, decimal=8)

    def test_hessian(self):
        x = np.array([1])
        y = np.array([2])
        Rs = 1.0
        sigma0 = 0.5
        f_xx, f_xy, f_yx, f_yy = self.profile.hessian(x, y, sigma0, Rs)
        npt.assert_almost_equal(f_xx[0], 0.0779016004481825, decimal=6)
        npt.assert_almost_equal(f_yy[0], -0.023212809452388683, decimal=6)
        npt.assert_almost_equal(f_xy[0], -0.0674096084507525, decimal=6)
        npt.assert_almost_equal(f_xy, f_yx, decimal=8)

    def test_mass_tot(self):
        rho0 = 1
        Rs = 3
        m_tot = self.profile.mass_tot(rho0, Rs)
        npt.assert_almost_equal(m_tot, 169.64600329384882, decimal=6)

    def test_grav_pot(self):
        x, y = 1, 0
        rho0 = 1
        Rs = 3
        grav_pot = self.profile.grav_pot(x, y, rho0, Rs, center_x=0, center_y=0)
        npt.assert_almost_equal(grav_pot, 42.411500823462205, decimal=8)

    def test_sigma0_definition(self):
        Rs = 2.0
        sigma0 = 0.5
        f_x, f_y = self.profile.derivatives(Rs, 0, sigma0, Rs)
        alpha = f_x
        npt.assert_almost_equal(alpha, 2 / 3.0 * sigma0 * Rs, decimal=5)


class TestHernquistEllipse(object):
    def setup_method(self):
        self.profile = Hernquist_Ellipse()

    def test_function(self):
        x = np.array([1])
        y = np.array([2])
        Rs = 1.0
        sigma0 = 0.5
        q, phi_G = 0.8, 0.5
        e1, e2 = param_util.phi_q2_ellipticity(phi_G, q)
        values = self.profile.function(x, y, sigma0, Rs, e1, e2)
        npt.assert_almost_equal(values[0], 0.6431256503990406, decimal=6)
        x = np.array([0])
        y = np.array([0])
        Rs = 1.0
        sigma0 = 0.5
        values = self.profile.function(x, y, sigma0, Rs, e1, e2)
        npt.assert_almost_equal(values[0], 0, decimal=6)

        x = np.array([2, 3, 4])
        y = np.array([1, 1, 1])
        values = self.profile.function(x, y, sigma0, Rs, e1, e2)
        npt.assert_almost_equal(values[0], 0.5983066087852915, decimal=6)
        npt.assert_almost_equal(values[1], 0.8014743038748537, decimal=6)
        npt.assert_almost_equal(values[2], 0.9806423979121521, decimal=6)

    def test_derivatives(self):
        x = 1
        y = 2
        Rs = 1.0
        sigma0 = 0.5
        q, phi_G = 0.8, 0.5
        e1, e2 = param_util.phi_q2_ellipticity(phi_G, q)
        f_x, f_y = self.profile.derivatives(x, y, sigma0, Rs, e1, e2)
        npt.assert_almost_equal(f_x, 0.06066651014748484, decimal=6)
        npt.assert_almost_equal(f_y, 0.24321226850316485, decimal=6)
        x = np.array([0])
        y = np.array([0])
        f_x, f_y = self.profile.derivatives(x, y, sigma0, Rs, e1, e2)
        npt.assert_almost_equal(f_x, 0, decimal=8)
        npt.assert_almost_equal(f_y, 0, decimal=8)

    def test_hessian(self):
        x = np.array([1.0])
        y = np.array([2.0])
        Rs = 1.0
        sigma0 = 0.5
        q, phi_G = 0.8, 0.5
        e1, e2 = param_util.phi_q2_ellipticity(phi_G, q)
        f_xx, f_xy, f_yx, f_yy = self.profile.hessian(x, y, sigma0, Rs, e1, e2)
        npt.assert_almost_equal(f_xx[0], 0.09437634437420293, decimal=6)
        npt.assert_almost_equal(f_yy[0], -0.029223594277816286, decimal=6)
        npt.assert_almost_equal(f_xy[0], -0.062239983771028566, decimal=6)
        npt.assert_almost_equal(f_xy, f_yx, decimal=6)


if __name__ == "__main__":
    pytest.main()



import numpy as np
from lenstronomy.LensModel.Profiles.base_profile import LensProfileBase

__all__ = ["Hernquist"]


class Hernquist(LensProfileBase):
    """Class to compute the Hernquist 1990 model, which is in 3d: rho(r) = rho0 / (r/Rs
    * (1 + (r/Rs))**3)

    in lensing terms, the normalization parameter 'sigma0' is defined such that the deflection at projected RS leads to
    alpha = 2./3 * Rs * sigma0

    Examples for converting angular to physical mass units
    ------------------------------------------------------

    >>> from lenstronomy.Cosmo.lens_cosmo import LensCosmo
    >>> from astropy.cosmology import FlatLambdaCDM
    >>> cosmo = FlatLambdaCDM(H0=70, Om0=0.3, Ob0=0.05)
    >>> lens_cosmo = LensCosmo(z_lens=0.5, z_source=1.5, cosmo=cosmo)

    Here we compute the angular scale of Rs on the sky (in arc seconds) and the deflection the normalization sigma0 from
    the total stellar mass in M_sol and Rs in [Mpc]:

    >>> sigma0, rs_angle = lens_cosmo.hernquist_phys2angular(mass=10**11, rs=0.02)

    And here we perform the inverse calculation given Rs_angle and alpha_Rs to return the physical halo properties.

    >>> m_tot, rs = lens_cosmo.hernquist_angular2phys(sigma0=sigma0 rs_angle=rs_angle)

    The lens model calculation uses angular units as arguments! So to execute a deflection angle calculation one uses

    >>> from lenstronomy.LensModel.Profiles.hernquist import Hernquist
    >>> hernquist = Hernquist()
    >>> alpha_x, alpha_y = hernquist.derivatives(x=1, y=1, Rs=rs_angle, sigma0=sigma0, center_x=0, center_y=0)
    """

    _diff = 0.00001
    _s = 0.00001
    param_names = ["sigma0", "Rs", "center_x", "center_y"]
    lower_limit_default = {"sigma0": 0, "Rs": 0, "center_x": -100, "center_y": -100}
    upper_limit_default = {"sigma0": 100, "Rs": 100, "center_x": 100, "center_y": 100}

    @staticmethod
    def density(r, rho0, Rs):
        """Computes the 3-d density.

        :param r: 3-d radius
        :param rho0: density normalization
        :param Rs: Hernquist radius
        :return: density at radius r
        """
        rho = rho0 / (r / Rs * (1 + (r / Rs)) ** 3)
        return rho

    def density_lens(self, r, sigma0, Rs):
        """Density as a function of 3d radius in lensing parameters This function
        converts the lensing definition sigma0 into the 3d density.

        :param r: 3d radius
        :param sigma0: rho0 * Rs (units of projected density)
        :param Rs: Hernquist radius
        :return: enclosed mass in 3d
        """
        rho0 = self.sigma2rho(sigma0, Rs)
        return self.density(r, rho0, Rs)

    def density_2d(self, x, y, rho0, Rs, center_x=0, center_y=0):
        """Projected density along the line of sight at coordinate (x, y)

        :param x: x-coordinate
        :param y: y-coordinate
        :param rho0: density normalization
        :param Rs: Hernquist radius
        :param center_x: x-center of the profile
        :param center_y: y-center of the profile
        :return: projected density
        """
        x_ = x - center_x
        y_ = y - center_y
        r = np.sqrt(x_**2 + y_**2)
        X = r / Rs
        sigma0 = self.rho2sigma(rho0, Rs)
        if isinstance(X, int) or isinstance(X, float):
            if X == 1:
                X = 1.000001
        else:
            X[X == 1] = 1.000001
        sigma = sigma0 / (X**2 - 1) ** 2 * (-3 + (2 + X**2) * self._F(X))
        return sigma

    @staticmethod
    def mass_3d(r, rho0, Rs):
        """Mass enclosed a 3d sphere or radius r.

        :param r: 3-d radius within the mass is integrated (same distance units as
            density definition)
        :param rho0: density normalization
        :param Rs: Hernquist radius
        :return: enclosed mass
        """
        mass_3d = 2 * np.pi * Rs**3 * rho0 * r**2 / (r + Rs) ** 2
        return mass_3d

    def mass_3d_lens(self, r, sigma0, Rs):
        """Mass enclosed a 3d sphere or radius r for lens parameterisation This function
        converts the lensing definition sigma0 into the 3d density.

        :param r: radius
        :param sigma0: rho0 * Rs (units of projected density)
        :param Rs: Hernquist radius
        :return: enclosed mass in 3d
        """
        rho0 = self.sigma2rho(sigma0, Rs)
        return self.mass_3d(r, rho0, Rs)

    def mass_2d(self, r, rho0, Rs):
        """Mass enclosed projected 2d sphere of radius r.

        :param r: projected radius
        :param rho0: density normalization
        :param Rs: Hernquist radius
        :return: mass enclosed 2d projected radius
        """

        sigma0 = self.rho2sigma(rho0, Rs)
        return self.mass_2d_lens(r, sigma0, Rs)

    def mass_2d_lens(self, r, sigma0, Rs):
        """Mass enclosed projected 2d sphere of radius r Same as mass_2d but with input
        normalization in units of projected density.

        :param r: projected radius
        :param sigma0: rho0 * Rs (units of projected density)
        :param Rs: Hernquist radius
        :return: mass enclosed 2d projected radius
        """
        X = r / Rs
        alpha_r = 2 * sigma0 * Rs * X * (1 - self._F(X)) / (X**2 - 1)
        mass_2d = alpha_r * r * np.pi
        return mass_2d

    @staticmethod
    def mass_tot(rho0, Rs):
        """Total mass within the profile.

        :param rho0: density normalization
        :param Rs: Hernquist radius
        :return: total mass within profile
        """
        m_tot = 2 * np.pi * rho0 * Rs**3
        return m_tot

    def function(self, x, y, sigma0, Rs, center_x=0, center_y=0):
        """Lensing potential.

        :param x: x-coordinate position (units of angle)
        :param y: y-coordinate position (units of angle)
        :param sigma0: normalization parameter defined such that the deflection at
            projected RS leads to alpha = 2./3 * Rs * sigma0
        :param Rs: Hernquist radius in units of angle
        :param center_x: x-center of the profile (units of angle)
        :param center_y: y-center of the profile (units of angle)
        :return: lensing potential at (x,y)
        """
        x_ = x - center_x
        y_ = y - center_y
        r = np.sqrt(x_**2 + y_**2)
        r = np.maximum(r, self._s)
        X = r / Rs
        f_ = sigma0 * Rs**2 * (np.log(X**2 / 4.0) + 2 * self._F(X))
        return f_

    def derivatives(self, x, y, sigma0, Rs, center_x=0, center_y=0):
        """

        :param x: x-coordinate position (units of angle)
        :param y: y-coordinate position (units of angle)
        :param sigma0: normalization parameter defined such that the deflection at
            projected RS leads to alpha = 2./3 * Rs * sigma0
        :param Rs: Hernquist radius in units of angle
        :param center_x: x-center of the profile (units of angle)
        :param center_y: y-center of the profile (units of angle)
        :return: derivative of function (deflection angles in x- and y-direction)
        """
        x_ = x - center_x
        y_ = y - center_y
        r = np.sqrt(x_**2 + y_**2)
        r = np.maximum(r, self._s)
        X = r / Rs
        if isinstance(r, int) or isinstance(r, float):
            # f = (1 - self._F(X)) / (X ** 2 - 1)  # this expression is 1/3 for X=1
            if X == 1:
                f = 1.0 / 3
            else:
                f = (1 - self._F(X)) / (X**2 - 1)
        else:
            f = np.empty_like(X)
            f[X == 1] = 1.0 / 3
            X_ = X[X != 1]
            f[X != 1] = (1 - self._F(X_)) / (X_**2 - 1)
        alpha_r = 2 * sigma0 * Rs * f * X
        f_x = alpha_r * x_ / r
        f_y = alpha_r * y_ / r
        return f_x, f_y

    def hessian(self, x, y, sigma0, Rs, center_x=0, center_y=0):
        """Hessian terms of the function.

        :param x: x-coordinate position (units of angle)
        :param y: y-coordinate position (units of angle)
        :param sigma0: normalization parameter defined such that the deflection at
            projected RS leads to alpha = 2./3 * Rs * sigma0
        :param Rs: Hernquist radius in units of angle
        :param center_x: x-center of the profile (units of angle)
        :param center_y: y-center of the profile (units of angle)
        :return: df/dxdx, df/dxdy, df/dydx, df/dydy
        """
        diff = self._diff
        alpha_ra_dx, alpha_dec_dx = self.derivatives(
            x + diff / 2, y, sigma0, Rs, center_x, center_y
        )
        alpha_ra_dy, alpha_dec_dy = self.derivatives(
            x, y + diff / 2, sigma0, Rs, center_x, center_y
        )

        alpha_ra_dx_, alpha_dec_dx_ = self.derivatives(
            x - diff / 2, y, sigma0, Rs, center_x, center_y
        )
        alpha_ra_dy_, alpha_dec_dy_ = self.derivatives(
            x, y - diff / 2, sigma0, Rs, center_x, center_y
        )

        f_xx = (alpha_ra_dx - alpha_ra_dx_) / diff
        f_xy = (alpha_ra_dy - alpha_ra_dy_) / diff
        f_yx = (alpha_dec_dx - alpha_dec_dx_) / diff
        f_yy = (alpha_dec_dy - alpha_dec_dy_) / diff
        return f_xx, f_xy, f_yx, f_yy

    @staticmethod
    def rho2sigma(rho0, Rs):
        """Converts 3d density into 2d projected density parameter.

        :param rho0: 3d density normalization of Hernquist model
        :param Rs: Hernquist radius
        :return: sigma0 defined quantity in projected units
        """
        return rho0 * Rs

    @staticmethod
    def sigma2rho(sigma0, Rs):
        """Converts projected density parameter (in units of deflection) into 3d density
        parameter.

        :param sigma0: density defined quantity in projected units
        :param Rs: Hernquist radius
        :return: rho0 the 3d density normalization of Hernquist model
        """
        return sigma0 / Rs

    def _F(self, X):
        """
        function 48 in https://arxiv.org/pdf/astro-ph/0102341.pdf
        :param X: r/rs
        :return: F(X)
        """
        c = self._s
        if isinstance(X, int) or isinstance(X, float):
            X = max(X, c)
            if 0 < X < 1:
                a = 1.0 / np.sqrt(1 - X**2) * np.arctanh(np.sqrt(1 - X**2))
            elif X == 1:
                a = 1.0
            elif X > 1:
                a = 1.0 / np.sqrt(X**2 - 1) * np.arctan(np.sqrt(X**2 - 1))
            else:  # X == 0:
                a = 1.0 / np.sqrt(1 - c**2) * np.arctanh(np.sqrt((1 - c**2)))

        else:
            a = np.empty_like(X)
            X[X < c] = c
            x = X[X < 1]
            a[X < 1] = 1 / np.sqrt(1 - x**2) * np.arctanh(np.sqrt((1 - x**2)))

            # x = X[X == 1]
            a[X == 1] = 1.0

            x = X[X > 1]
            a[X > 1] = 1 / np.sqrt(x**2 - 1) * np.arctan(np.sqrt(x**2 - 1))
            # a[X>y] = 0
        return a

    def grav_pot(self, x, y, rho0, Rs, center_x=0, center_y=0):
        """#TODO decide whether these functions are needed or not

        gravitational potential (modulo 4 pi G and rho0 in appropriate units)
        :param x: x-coordinate position (units of angle)
        :param y: y-coordinate position (units of angle)
        :param rho0: density normalization parameter of Hernquist profile
        :param Rs: Hernquist radius in units of angle
        :param center_x: x-center of the profile (units of angle)
        :param center_y: y-center of the profile (units of angle)
        :return: gravitational potential at projected radius
        """
        x_ = x - center_x
        y_ = y - center_y
        r = np.sqrt(x_**2 + y_**2)
        M = self.mass_tot(rho0, Rs)
        pot = M / (r + Rs)
        return pot



__author__ = "herjy", "aymgal", "sibirrer"

import numpy as np
from scipy import ndimage

from lenstronomy.Util.package_util import exporter

export, __all__ = exporter()


@export
def transform(img, n_scales, second_gen=False):
    """Performs starlet decomposition of an 2D array.

    :param img: input image
    :param n_scales: number of decomposition scales
    :param second_gen: if True, 'second generation' starlets are used
    """
    mode = "nearest"

    lvl = n_scales - 1
    sh = np.shape(img)

    n1 = sh[1]
    n2 = sh[1]

    # B-spline filter
    h = [1.0 / 16, 1.0 / 4, 3.0 / 8, 1.0 / 4, 1.0 / 16]
    n = np.size(h)
    h = np.array(h)

    max_lvl = np.min((lvl, int(np.log2(n2))))
    if lvl > max_lvl:
        raise ValueError(
            "Maximum decomposition level is {} (required: {})".format(max_lvl, lvl)
        )
    elif lvl <= 0:
        raise ValueError("Number of decomposition level can not be non-positive")

    c = img
    # wavelet set of coefficients.
    wave = np.zeros((lvl + 1, n1, n2))

    for i in range(lvl):
        newh = np.zeros((1, n + (n - 1) * (2**i - 1)))
        newh[0, np.linspace(0, np.size(newh) - 1, len(h), dtype=int)] = h

        # H = np.dot(newh.T, newh)

        ######Calculates c(j+1)
        ###### Line convolution
        cnew = ndimage.convolve1d(c, newh[0, :], axis=0, mode=mode)

        ###### Column convolution
        cnew = ndimage.convolve1d(cnew, newh[0, :], axis=1, mode=mode)

        if second_gen:
            ###### hoh for g; Column convolution
            hc = ndimage.convolve1d(cnew, newh[0, :], axis=0, mode=mode)

            ###### hoh for g; Line convolution
            hc = ndimage.convolve1d(hc, newh[0, :], axis=1, mode=mode)

            ###### wj+1 = cj - hcj+1
            wave[i, :, :] = c - hc

        else:
            ###### wj+1 = cj - cj+1
            wave[i, :, :] = c - cnew

        c = cnew

    wave[i + 1, :, :] = c

    return wave


@export
def inverse_transform(wave, fast=True, second_gen=False):
    """Reconstructs an image fron its starlet decomposition coefficients.

    :param wave: input coefficients, with shape (n_scales, np.sqrt(n_pixel),
        np.sqrt(n_pixel))
    :param fast: if True, and only with second_gen is False, simply sums up all scales
        to reconstruct the image
    :param second_gen: if True, 'second generation' starlets are used
    """
    if fast and not second_gen:
        # simply sum all scales, including the coarsest one
        return np.sum(wave, axis=0)

    mode = "nearest"

    lvl, n1, n2 = np.shape(wave)
    h = np.array([1.0 / 16, 1.0 / 4, 3.0 / 8, 1.0 / 4, 1.0 / 16])
    n = np.size(h)

    cJ = np.copy(wave[lvl - 1, :, :])

    for i in range(1, lvl):
        newh = np.zeros((1, n + (n - 1) * (2 ** (lvl - 1 - i) - 1)))
        newh[0, np.linspace(0, np.size(newh) - 1, len(h), dtype=int)] = h
        H = np.dot(newh.T, newh)

        ###### Line convolution
        cnew = ndimage.convolve1d(cJ, newh[0, :], axis=0, mode=mode)
        ###### Column convolution
        cnew = ndimage.convolve1d(cnew, newh[0, :], axis=1, mode=mode)

        cJ = cnew + wave[lvl - 1 - i, :, :]

    return np.reshape(cJ, (n1, n2))



__author__ = "sibirrer"
import pytest
import numpy as np
import numpy.testing as npt

import lenstronomy.Util.util as util
from lenstronomy.LensModel.Profiles.sis import SIS
from lenstronomy.LensModel.Profiles.interpol import Interpol, InterpolScaled


class TestInterpol(object):
    def setup_method(self):
        pass

    def test_do_interpol(self):
        numPix = 101
        deltaPix = 0.1
        x_grid_interp, y_grid_interp = util.make_grid(numPix, deltaPix)
        sis = SIS()
        kwargs_SIS = {"theta_E": 1.0, "center_x": 0.5, "center_y": -0.5}
        f_sis = sis.function(x_grid_interp, y_grid_interp, **kwargs_SIS)
        f_x_sis, f_y_sis = sis.derivatives(x_grid_interp, y_grid_interp, **kwargs_SIS)
        f_xx_sis, f_xy_sis, f_yx_sis, f_yy_sis = sis.hessian(
            x_grid_interp, y_grid_interp, **kwargs_SIS
        )
        x_axes, y_axes = util.get_axes(x_grid_interp, y_grid_interp)
        interp_func = Interpol(grid=True)
        interp_func_loop = Interpol(grid=False)
        interp_func.do_interp(
            x_axes,
            y_axes,
            util.array2image(f_sis),
            util.array2image(f_x_sis),
            util.array2image(f_y_sis),
            util.array2image(f_xx_sis),
            util.array2image(f_yy_sis),
            util.array2image(f_xy_sis),
        )
        interp_func_loop.do_interp(
            x_axes,
            y_axes,
            util.array2image(f_sis),
            util.array2image(f_x_sis),
            util.array2image(f_y_sis),
            util.array2image(f_xx_sis),
            util.array2image(f_yy_sis),
            util.array2image(f_xy_sis),
        )

        # test derivatives
        print(interp_func.derivatives(0, 1))
        print(sis.derivatives(1, 0, **kwargs_SIS))
        # assert interp_func.derivatives(1, 0) == sis.derivatives(1, 0, **kwargs_SIS)
        assert interp_func.derivatives(1, 0) == interp_func_loop.derivatives(1, 0)
        alpha1_interp, alpha2_interp = interp_func.derivatives(
            np.array([0, 1, 0, 1]), np.array([1, 1, 2, 2])
        )
        alpha1_interp_loop, alpha2_interp_loop = interp_func_loop.derivatives(
            np.array([0, 1, 0, 1]), np.array([1, 1, 2, 2])
        )
        alpha1_true, alpha2_true = sis.derivatives(
            np.array([0, 1, 0, 1]), np.array([1, 1, 2, 2]), **kwargs_SIS
        )
        assert alpha1_interp[0] == alpha1_true[0]
        assert alpha1_interp[1] == alpha1_true[1]
        assert alpha1_interp[0] == alpha1_interp_loop[0]
        assert alpha1_interp[1] == alpha1_interp_loop[1]
        # test hessian
        assert interp_func.hessian(1, 0) == sis.hessian(1, 0, **kwargs_SIS)
        f_xx_interp, f_xy_interp, f_yx_interp, f_yy_interp = interp_func.hessian(
            np.array([0, 1, 0, 1]), np.array([1, 1, 2, 2])
        )
        (
            f_xx_interp_loop,
            f_xy_interp_loop,
            f_yx_interp_loop,
            f_yy_interp_loop,
        ) = interp_func_loop.hessian(np.array([0, 1, 0, 1]), np.array([1, 1, 2, 2]))
        f_xx_true, f_xy_true, f_yx_true, f_yy_true = sis.hessian(
            np.array([0, 1, 0, 1]), np.array([1, 1, 2, 2]), **kwargs_SIS
        )
        assert f_xx_interp[0] == f_xx_true[0]
        assert f_xx_interp[1] == f_xx_true[1]
        assert f_xy_interp[0] == f_xy_true[0]
        assert f_xy_interp[1] == f_xy_true[1]
        assert f_xx_interp[0] == f_xx_interp_loop[0]
        assert f_xx_interp[1] == f_xx_interp_loop[1]
        assert f_xy_interp[0] == f_xy_interp_loop[0]
        assert f_xy_interp[1] == f_xy_interp_loop[1]
        # test all

    def test_call(self):
        numPix = 101
        deltaPix = 0.1
        x_grid_interp, y_grid_interp = util.make_grid(numPix, deltaPix)
        sis = SIS()
        kwargs_SIS = {"theta_E": 1.0, "center_x": 0.5, "center_y": -0.5}
        f_sis = sis.function(x_grid_interp, y_grid_interp, **kwargs_SIS)
        f_x_sis, f_y_sis = sis.derivatives(x_grid_interp, y_grid_interp, **kwargs_SIS)
        f_xx_sis, f_xy_sis, f_yx_sis, f_yy_sis = sis.hessian(
            x_grid_interp, y_grid_interp, **kwargs_SIS
        )
        x_axes, y_axes = util.get_axes(x_grid_interp, y_grid_interp)
        interp_func = Interpol(grid=True)
        interp_func.do_interp(
            x_axes,
            y_axes,
            util.array2image(f_sis),
            util.array2image(f_x_sis),
            util.array2image(f_y_sis),
            util.array2image(f_xx_sis),
            util.array2image(f_yy_sis),
            util.array2image(f_xy_sis),
        )
        x, y = 1.0, 1.0
        alpha_x, alpha_y = interp_func.derivatives(x, y, **{})
        assert alpha_x == 0.31622776601683794

    def test_kwargs_interpolation(self):
        numPix = 101
        deltaPix = 0.1
        x_grid_interp, y_grid_interp = util.make_grid(numPix, deltaPix)
        sis = SIS()
        kwargs_SIS = {"theta_E": 1.0, "center_x": 0.5, "center_y": -0.5}
        f_sis = sis.function(x_grid_interp, y_grid_interp, **kwargs_SIS)
        f_x_sis, f_y_sis = sis.derivatives(x_grid_interp, y_grid_interp, **kwargs_SIS)
        f_xx_sis, f_xy_sis, f_yx_sis, f_yy_sis = sis.hessian(
            x_grid_interp, y_grid_interp, **kwargs_SIS
        )
        x_axes, y_axes = util.get_axes(x_grid_interp, y_grid_interp)
        interp_func = Interpol()
        kwargs_interp = {
            "grid_interp_x": x_axes,
            "grid_interp_y": y_axes,
            "f_": util.array2image(f_sis),
            "f_x": util.array2image(f_x_sis),
            "f_y": util.array2image(f_y_sis),
            "f_xx": util.array2image(f_xx_sis),
            "f_yy": util.array2image(f_yy_sis),
            "f_xy": util.array2image(f_xy_sis),
        }
        x, y = 1.0, 1.0
        alpha_x, alpha_y = interp_func.derivatives(x, y, **kwargs_interp)
        assert alpha_x == 0.31622776601683794

        x, y = 1.0, 0.0
        alpha_x, alpha_y = interp_func.derivatives(x, y, **kwargs_interp)
        alpha_x_true, alpha_y_true = sis.derivatives(x, y, **kwargs_SIS)
        npt.assert_almost_equal(alpha_x, alpha_x_true, decimal=10)
        npt.assert_almost_equal(alpha_y, alpha_y_true, decimal=10)

        f_ = interp_func.function(x, y, **kwargs_interp)
        f_true = sis.derivatives(x, y, **kwargs_SIS)
        npt.assert_almost_equal(f_, f_true, decimal=10)

    def test_hessian_finite_differential(self):
        numPix = 101
        deltaPix = 0.1
        x_grid_interp, y_grid_interp = util.make_grid(numPix, deltaPix)
        sis = SIS()
        kwargs_SIS = {"theta_E": 1.0, "center_x": 0.5, "center_y": -0.5}
        f_sis = sis.function(x_grid_interp, y_grid_interp, **kwargs_SIS)
        f_x_sis, f_y_sis = sis.derivatives(x_grid_interp, y_grid_interp, **kwargs_SIS)
        x_axes, y_axes = util.get_axes(x_grid_interp, y_grid_interp)
        interp_func = Interpol()
        kwargs_interp = {
            "grid_interp_x": x_axes,
            "grid_interp_y": y_axes,
            "f_": util.array2image(f_sis),
            "f_x": util.array2image(f_x_sis),
            "f_y": util.array2image(f_y_sis),
        }
        x, y = 1.0, 0.0
        f_xx, f_xy, f_yx, f_yy = interp_func.hessian(x, y, **kwargs_interp)
        f_xx_true, f_xy_true, f_yx_true, f_yy_true = sis.hessian(x, y, **kwargs_SIS)
        npt.assert_almost_equal(f_xx, f_xx_true, decimal=1)
        npt.assert_almost_equal(f_xy, f_xy_true, decimal=1)
        npt.assert_almost_equal(f_yx, f_yx_true, decimal=1)
        npt.assert_almost_equal(f_yy, f_yy_true, decimal=1)

    def test_interp_func_scaled(self):
        numPix = 101
        deltaPix = 0.1
        x_grid_interp, y_grid_interp = util.make_grid(numPix, deltaPix)
        sis = SIS()
        kwargs_SIS = {"theta_E": 1.0, "center_x": 0.5, "center_y": -0.5}
        f_sis = sis.function(x_grid_interp, y_grid_interp, **kwargs_SIS)
        f_x_sis, f_y_sis = sis.derivatives(x_grid_interp, y_grid_interp, **kwargs_SIS)
        f_xx_sis, f_xy_sis, f_yx_sis, f_yy_sis = sis.hessian(
            x_grid_interp, y_grid_interp, **kwargs_SIS
        )
        x_axes, y_axes = util.get_axes(x_grid_interp, y_grid_interp)
        kwargs_interp = {
            "grid_interp_x": x_axes,
            "grid_interp_y": y_axes,
            "f_": util.array2image(f_sis),
            "f_x": util.array2image(f_x_sis),
            "f_y": util.array2image(f_y_sis),
            "f_xx": util.array2image(f_xx_sis),
            "f_yy": util.array2image(f_yy_sis),
            "f_xy": util.array2image(f_xy_sis),
        }
        interp_func = InterpolScaled(grid=False)
        x, y = 1.0, 1.0
        alpha_x, alpha_y = interp_func.derivatives(
            x, y, scale_factor=1, **kwargs_interp
        )
        assert alpha_x == 0.31622776601683794

        f_ = interp_func.function(x, y, scale_factor=1.0, **kwargs_interp)
        npt.assert_almost_equal(f_, 1.5811388300841898)

        f_xx, f_xy, f_yx, f_yy = interp_func.hessian(
            x, y, scale_factor=1.0, **kwargs_interp
        )
        npt.assert_almost_equal(f_xx, 0.56920997883030822, decimal=8)
        npt.assert_almost_equal(f_yy, 0.063245553203367583, decimal=8)
        npt.assert_almost_equal(f_xy, -0.18973665961010275, decimal=8)
        npt.assert_almost_equal(f_xy, f_yx, decimal=8)

        x_grid, y_grid = util.make_grid(10, deltaPix)
        f_xx, f_xy, f_yx, f_yy = interp_func.hessian(
            x_grid, y_grid, scale_factor=1.0, **kwargs_interp
        )
        npt.assert_almost_equal(f_xx[0], 0, decimal=2)

    def test_shift(self):
        numPix = 101
        deltaPix = 0.1
        x_grid_interp, y_grid_interp = util.make_grid(numPix, deltaPix)
        sis = SIS()

        kwargs_SIS = {"theta_E": 1.0, "center_x": 0.5, "center_y": -0.5}
        f_sis = sis.function(x_grid_interp, y_grid_interp, **kwargs_SIS)
        f_x_sis, f_y_sis = sis.derivatives(x_grid_interp, y_grid_interp, **kwargs_SIS)
        f_xx_sis, f_xy_sis, f_yx_sis, f_yy_sis = sis.hessian(
            x_grid_interp, y_grid_interp, **kwargs_SIS
        )
        x_axes, y_axes = util.get_axes(x_grid_interp, y_grid_interp)
        kwargs_interp = {
            "grid_interp_x": x_axes,
            "grid_interp_y": y_axes,
            "f_": util.array2image(f_sis),
            "f_x": util.array2image(f_x_sis),
            "f_y": util.array2image(f_y_sis),
            "f_xx": util.array2image(f_xx_sis),
            "f_yy": util.array2image(f_yy_sis),
            "f_xy": util.array2image(f_xy_sis),
        }
        interp_func = Interpol(grid=False)
        x, y = 1.0, 1.0
        alpha_x, alpha_y = interp_func.derivatives(x, y, **kwargs_interp)
        assert alpha_x == 0.31622776601683794

        interp_func = Interpol(grid=False)
        x_shift = 1.0
        kwargs_shift = {
            "grid_interp_x": x_axes + x_shift,
            "grid_interp_y": y_axes,
            "f_": util.array2image(f_sis),
            "f_x": util.array2image(f_x_sis),
            "f_y": util.array2image(f_y_sis),
            "f_xx": util.array2image(f_xx_sis),
            "f_yy": util.array2image(f_yy_sis),
            "f_xy": util.array2image(f_xy_sis),
        }
        alpha_x_shift, alpha_y_shift = interp_func.derivatives(
            x + x_shift, y, **kwargs_shift
        )
        npt.assert_almost_equal(alpha_x_shift, alpha_x, decimal=10)


if __name__ == "__main__":
    pytest.main()



__author__ = "sibirrer"
"""This class contains physical constants and conversion factors between units."""
import numpy as np

__all__ = (
    "G c M_sun M_earth M_jupiter AU Mpc kpc pc day_s arcsec "
    "a_ES F_ES delay_arcsec2days".split()
)

G = 6.67384 * 10 ** (-11)  # Gravitational constant [m^3 kg^-1 s^-2]
c = 299792458  # [m/s]

M_sun = 1.9891 * 10**30  # solar mass in [kg]
M_earth = 5.9972 * 10**24  # Earth mass in [kg]
M_jupiter = 1.89813 * 10**27  # Jupiter mass in [kg]
AU = 1.495978707 * 10**11  # Distance Earth Sun (Astronomical unit) in [m]

Mpc = 3.08567758 * 10**22  # Mpc in [m]
kpc = 3.08567758 * 10**19  # kpc in [m]
pc = 3.08567758 * 10**16  # pc in [m]
r_sun = 6.96000 * 10**8  # radius of the sun in [m]

day_s = 24 * 3600  # day in second
arcsec = 2 * np.pi / 360 / 3600  # arc second in radian

# derived quantities

a_ES = G * M_sun / AU**2  # Earth-Sun acceleration
F_ES = G * M_sun * M_earth / AU**2


def delay_arcsec2days(delay_arcsec, ddt):
    """Given a delay in arcsec^2 and a Delay distance, the delay is computed in days.

    :param delay_arcsec: gravitational delay in units of arcsec^2 (e.g. Fermat
        potential)
    :param ddt: Time delay distance (in units of Mpc)
    :return: time-delay in units of days
    """
    return ddt * Mpc / c * delay_arcsec / day_s * arcsec**2



import pytest
from lenstronomy.GalKin.galkin_multiobservation import GalkinMultiObservation


class TestGalkinMultiObservation(object):
    def setup_method(self):
        pass

    def test_dispersion(self):
        light_profile_list = ["HERNQUIST"]
        Rs = 0.5
        kwargs_light = [
            {"Rs": Rs, "amp": 1.0}
        ]  # effective half light radius (2d projected) in arcsec
        # 0.551 *
        # mass profile
        mass_profile_list = ["SPP"]
        theta_E = 1.2
        gamma = 2.0
        kwargs_mass = [
            {"theta_E": theta_E, "gamma": gamma}
        ]  # Einstein radius (arcsec) and power-law slope

        # anisotropy profile
        anisotropy_type = "OM"
        r_ani = 2.0
        kwargs_anisotropy = {"r_ani": r_ani}  # anisotropy radius [arcsec]

        kwargs_model = {
            "mass_profile_list": mass_profile_list,
            "light_profile_list": light_profile_list,
            "anisotropy_model": anisotropy_type,
        }
        kwargs_cosmo = {"d_d": 1000, "d_s": 1500, "d_ds": 800}
        kwargs_numerics = {
            "interpol_grid_num": 500,
            "log_integration": True,
            "max_integrate": 10,
            "min_integrate": 0.001,
        }

        # aperture as slit
        aperture_type = "slit"
        kwargs_aperture_1 = {"width": 1, "length": 1.0, "aperture_type": aperture_type}
        kwargs_psf_1 = {"psf_type": "GAUSSIAN", "fwhm": 0.7}

        kwargs_aperture_2 = {"width": 3, "length": 3.0, "aperture_type": aperture_type}
        kwargs_psf_2 = {"psf_type": "GAUSSIAN", "fwhm": 1.5}
        kwargs_aperture_list = [kwargs_aperture_1, kwargs_aperture_2]
        kwargs_psf_list = [kwargs_psf_1, kwargs_psf_2]
        galkin_multiobs = GalkinMultiObservation(
            kwargs_model,
            kwargs_aperture_list,
            kwargs_psf_list,
            kwargs_cosmo,
            kwargs_numerics=kwargs_numerics,
            analytic_kinematics=False,
        )

        sigma_v_list = galkin_multiobs.dispersion_map(
            kwargs_mass=kwargs_mass,
            kwargs_light=kwargs_light,
            kwargs_anisotropy=kwargs_anisotropy,
            num_kin_sampling=1000,
            num_psf_sampling=100,
        )
        assert len(sigma_v_list) == 2
        assert sigma_v_list[0] > sigma_v_list[1]


if __name__ == "__main__":
    pytest.main()



import pytest
import numpy as np
import numpy.testing as npt
from lenstronomy.LightModel.Profiles.nie import NIE
from lenstronomy.LightModel.Profiles.chameleon import (
    Chameleon,
    DoubleChameleon,
    TripleChameleon,
)
from lenstronomy.LensModel.Profiles.chameleon import Chameleon as ChameleonLens
import lenstronomy.Util.param_util as param_util
from lenstronomy.Util import util


class TestChameleon(object):
    """Class to test the Moffat profile."""

    def setup_method(self):
        pass

    def test_param_name(self):
        chameleon = Chameleon()
        names = chameleon.param_names
        assert names[0] == "amp"

    def test_function(self):
        """

        :return:
        """
        chameleon = Chameleon()
        nie = NIE()

        x = np.linspace(0.1, 10, 10)
        w_c, w_t = 0.5, 1.0
        phi_G, q = 0.3, 0.8
        e1, e2 = param_util.phi_q2_ellipticity(phi_G, q)
        kwargs_light = {"amp": 1.0, "w_c": 0.5, "w_t": 1.0, "e1": e1, "e2": e2}
        (
            amp_new,
            w_c,
            w_t,
            s_scale_1,
            s_scale_2,
        ) = chameleon._chameleonLens.param_convert(1, w_c, w_t, e1, e2)
        kwargs_1 = {"amp": amp_new, "s_scale": s_scale_1, "e1": e1, "e2": e2}
        kwargs_2 = {"amp": amp_new, "s_scale": s_scale_2, "e1": e1, "e2": e2}
        flux = chameleon.function(x=x, y=1.0, **kwargs_light)
        flux1 = nie.function(x=x, y=1.0, **kwargs_1)
        flux2 = nie.function(x=x, y=1.0, **kwargs_2)
        npt.assert_almost_equal(flux, flux1 - flux2, decimal=5)

    def test_lens_model_correspondence(self):
        """Here we test the proportionality of the convergence of the lens model with
        the surface brightness of the light model."""
        chameleon_lens = ChameleonLens()
        chameleon = Chameleon()

        x, y = util.make_grid(numPix=100, deltapix=0.1)
        e1, e2 = 0.0, 0
        w_c, w_t = 0.5, 1.0
        kwargs_light = {"amp": 1.0, "w_c": w_c, "w_t": w_t, "e1": e1, "e2": e2}
        kwargs_lens = {"alpha_1": 1.0, "w_c": w_c, "w_t": w_t, "e1": e1, "e2": e2}
        flux = chameleon.function(x=x, y=y, **kwargs_light)
        f_xx, f_xy, f_yx, f_yy = chameleon_lens.hessian(x=x, y=y, **kwargs_lens)
        kappa = 1 / 2.0 * (f_xx + f_yy)

        # flux2d = util.array2image(flux)
        # kappa2d = util.array2image(kappa)

        npt.assert_almost_equal(flux / np.mean(flux), kappa / np.mean(kappa), decimal=3)


class TestDoubleChameleon(object):
    """Class to test the Moffat profile."""

    def setup_method(self):
        pass

    def test_param_name(self):
        chameleon = DoubleChameleon()
        names = chameleon.param_names
        assert names[0] == "amp"

    def test_function(self):
        """

        :return:
        """
        chameleon = Chameleon()
        doublechameleon = DoubleChameleon()

        x = np.linspace(0.1, 10, 10)
        phi_G, q = 0.3, 0.8
        ratio = 2.0
        e1, e2 = param_util.phi_q2_ellipticity(phi_G, q)
        kwargs_light = {
            "amp": 1.0,
            "ratio": 2,
            "w_c1": 0.5,
            "w_t1": 1.0,
            "e11": e1,
            "e21": e2,
            "w_c2": 0.1,
            "w_t2": 0.5,
            "e12": e1,
            "e22": e2,
        }

        kwargs_1 = {
            "amp": 1.0 / (1 + 1.0 / ratio),
            "w_c": 0.5,
            "w_t": 1.0,
            "e1": e1,
            "e2": e2,
        }
        kwargs_2 = {
            "amp": 1.0 / (1 + ratio),
            "w_c": 0.1,
            "w_t": 0.5,
            "e1": e1,
            "e2": e2,
        }
        flux = doublechameleon.function(x=x, y=1.0, **kwargs_light)
        flux1 = chameleon.function(x=x, y=1.0, **kwargs_1)
        flux2 = chameleon.function(x=x, y=1.0, **kwargs_2)
        npt.assert_almost_equal(flux, flux1 + flux2, decimal=8)


class TestTripleChameleon(object):
    """Class to test the Moffat profile."""

    def setup_method(self):
        pass

    def test_param_name(self):
        chameleon = DoubleChameleon()
        names = chameleon.param_names
        assert names[0] == "amp"

    def test_function(self):
        """

        :return:
        """
        chameleon = Chameleon()
        triplechameleon = TripleChameleon()

        x = np.linspace(0.1, 10, 10)
        phi_G, q = 0.3, 0.8
        ratio12 = 2.0
        ratio13 = 3
        e1, e2 = param_util.phi_q2_ellipticity(phi_G, q)
        kwargs_light = {
            "amp": 1.0,
            "ratio12": ratio12,
            "ratio13": ratio13,
            "w_c1": 0.5,
            "w_t1": 1.0,
            "e11": e1,
            "e21": e2,
            "w_c2": 0.1,
            "w_t2": 0.5,
            "e12": e1,
            "e22": e2,
            "w_c3": 0.4,
            "w_t3": 0.8,
            "e13": e1,
            "e23": e2,
        }

        amp1 = 1.0 / (1.0 + 1.0 / ratio12 + 1.0 / ratio13)
        amp2 = amp1 / ratio12
        amp3 = amp1 / ratio13
        kwargs_1 = {"amp": amp1, "w_c": 0.5, "w_t": 1.0, "e1": e1, "e2": e2}
        kwargs_2 = {"amp": amp2, "w_c": 0.1, "w_t": 0.5, "e1": e1, "e2": e2}
        kwargs_3 = {"amp": amp3, "w_c": 0.4, "w_t": 0.8, "e1": e1, "e2": e2}
        flux = triplechameleon.function(x=x, y=1.0, **kwargs_light)
        flux1 = chameleon.function(x=x, y=1.0, **kwargs_1)
        flux2 = chameleon.function(x=x, y=1.0, **kwargs_2)
        flux3 = chameleon.function(x=x, y=1.0, **kwargs_3)
        npt.assert_almost_equal(flux, flux1 + flux2 + flux3, decimal=8)


if __name__ == "__main__":
    pytest.main()



import copy

from lenstronomy.Analysis.multi_patch_reconstruction import MultiPatchReconstruction
from lenstronomy.Plots import plot_util

import matplotlib.pyplot as plt
import numpy as np
from mpl_toolkits.axes_grid1 import make_axes_locatable


class MultiPatchPlot(MultiPatchReconstruction):
    """This class illustrates the model of disconnected multi-patch modeling with
    'joint-linear' option in one single array."""

    def __init__(
        self,
        multi_band_list,
        kwargs_model,
        kwargs_params,
        multi_band_type="joint-linear",
        kwargs_likelihood=None,
        kwargs_pixel_grid=None,
        verbose=True,
        cmap_string="gist_heat",
    ):
        """

        :param multi_band_list: list of imaging data configuration [[kwargs_data, kwargs_psf, kwargs_numerics], [...]]
        :param kwargs_model: model keyword argument list
        :param kwargs_params: keyword arguments of the model parameters, same as output of FittingSequence() 'kwargs_result'
        :param multi_band_type: string, option when having multiple imaging data sets modelled simultaneously. Options are:
            - 'multi-linear': linear amplitudes are inferred on single data set
            - 'linear-joint': linear amplitudes ae jointly inferred
            - 'single-band': single band
        :param kwargs_likelihood: likelihood keyword arguments as supported by the Likelihood() class
        :param kwargs_pixel_grid: keyword argument of PixelGrid() class. This is optional and overwrites a minimal grid
         Attention for consistent pixel grid definitions!
        :param verbose: if True (default), computes and prints the total log-likelihood.
        This can deactivated for speedup purposes (does not run linear inversion again), and reduces the number of prints.
        :param cmap_string: string of color map (or cmap matplotlib object)
        """
        MultiPatchReconstruction.__init__(
            self,
            multi_band_list,
            kwargs_model,
            kwargs_params,
            multi_band_type=multi_band_type,
            kwargs_likelihood=kwargs_likelihood,
            kwargs_pixel_grid=kwargs_pixel_grid,
            verbose=verbose,
        )
        (
            self._image_joint,
            self._model_joint,
            self._norm_residuals_joint,
        ) = self.image_joint()
        (
            self._kappa_joint,
            self._magnification_joint,
            self._alpha_x_joint,
            self._alpha_y_joint,
        ) = self.lens_model_joint()

        log_model = np.log10(self._model_joint)
        log_model[np.isnan(log_model)] = -5
        self._v_min_default = max(np.min(log_model), -5)
        self._v_max_default = min(np.max(log_model), 10)
        self._cmap = plot_util.cmap_conf(cmap_string)

    def data_plot(
        self,
        ax,
        log_scale=True,
        text="Observed",
        colorbar_label=r"log$_{10}$ flux",
        **kwargs
    ):
        """Illustrates data.

        :param ax: matplotlib axis instance
        :param kwargs: plotting keyword arguments
        :return: matplotlib instance
        """
        return self._plot(
            ax,
            image=self._image_joint,
            coords=self._pixel_grid_joint,
            log_scale=log_scale,
            text=text,
            colorbar_label=colorbar_label,
            **kwargs
        )

    def model_plot(
        self,
        ax,
        log_scale=True,
        text="Reconstructed",
        colorbar_label=r"log$_{10}$ flux",
        **kwargs
    ):
        """Illustrates model.

        :param ax: matplotlib axis instance
        :param kwargs: plotting keyword arguments
        :return: matplotlib instance
        """
        return self._plot(
            ax,
            image=self._model_joint,
            coords=self._pixel_grid_joint,
            log_scale=log_scale,
            text=text,
            colorbar_label=colorbar_label,
            **kwargs
        )

    def source_plot(
        self,
        ax,
        delta_pix,
        num_pix,
        center=None,
        log_scale=True,
        text="Source",
        colorbar_label=r"log$_{10}$ flux",
        dist_scale=0.1,
        **kwargs
    ):
        """Illustrates source.

        :param ax: matplotlib axis instance :param delta_pix scale of the pixel size of
            the source plot
        :param num_pix: number of pixels per axis of the source plot
        :param center: list with two entries [center_x, center_y] (optional)
        :param kwargs: plotting keyword arguments
        :return: matplotlib instance
        """
        source, coords = self.source(
            num_pix=num_pix, delta_pix=delta_pix, center=center
        )
        return self._plot(
            ax,
            image=source,
            coords=coords,
            log_scale=log_scale,
            text=text,
            colorbar_label=colorbar_label,
            dist_scale=dist_scale,
            **kwargs
        )

    def normalized_residual_plot(
        self,
        ax,
        v_min=-6,
        v_max=6,
        log_scale=False,
        text="Normalized Residuals",
        colorbar_label=r"(f${}_{\rm model}$ - f${}_{\rm data}$)/$\sigma$",
        cmap="bwr",
        white_on_black=False,
        **kwargs
    ):
        """
        illustrates normalized residuals of (data - model) / error

        :param ax: matplotlib axis instance
        :param kwargs: plotting keyword arguments
        :return: matplotlib instance
        """
        return self._plot(
            ax,
            image=self._norm_residuals_joint,
            coords=self._pixel_grid_joint,
            v_min=v_min,
            v_max=v_max,
            log_scale=log_scale,
            text=text,
            colorbar_label=colorbar_label,
            cmap=cmap,
            white_on_black=white_on_black,
            **kwargs
        )

    def convergence_plot(
        self,
        ax,
        log_scale=True,
        v_min=-2,
        v_max=0.2,
        text="Convergence",
        colorbar_label=r"$\log_{10}\ \kappa$",
        **kwargs
    ):
        """Illustrates lensing convergence.

        :param ax: matplotlib axis instance
        :param kwargs: plotting keyword arguments
        :return: matplotlib instance
        """
        return self._plot(
            ax,
            image=self._kappa_joint,
            coords=self._pixel_grid_joint,
            log_scale=log_scale,
            v_min=v_min,
            v_max=v_max,
            text=text,
            colorbar_label=colorbar_label,
            **kwargs
        )

    def magnification_plot(
        self,
        ax,
        log_scale=False,
        v_min=-10,
        v_max=10,
        text="Magnification",
        colorbar_label=r"$\det\ (\mathsf{A}^{-1})$",
        cmap="bwr",
        white_on_black=False,
        **kwargs
    ):
        """Illustrates lensing convergence.

        :param ax: matplotlib axis instance
        :param kwargs: plotting keyword arguments
        :return: matplotlib instance
        """
        return self._plot(
            ax,
            image=self._magnification_joint,
            coords=self._pixel_grid_joint,
            log_scale=log_scale,
            v_min=v_min,
            v_max=v_max,
            text=text,
            colorbar_label=colorbar_label,
            cmap=cmap,
            white_on_black=white_on_black,
            **kwargs
        )

    def plot_main(self, **kwargs):
        """Print the main plots together in a joint frame.

        :return:
        """

        f, axes = plt.subplots(2, 3, figsize=(16, 8))
        self.data_plot(ax=axes[0, 0], **kwargs)
        self.model_plot(ax=axes[0, 1], image_names=True, **kwargs)
        kwargs_residuals = copy.deepcopy(kwargs)
        if "v_min" in kwargs_residuals:
            kwargs_residuals.pop("v_min")
        if "v_max" in kwargs_residuals:
            kwargs_residuals.pop("v_max")
        self.normalized_residual_plot(
            ax=axes[0, 2], v_min=-6, v_max=6, **kwargs_residuals
        )
        self.source_plot(ax=axes[1, 0], delta_pix=0.01, num_pix=100, **kwargs)
        self.convergence_plot(ax=axes[1, 1], **kwargs)
        self.magnification_plot(ax=axes[1, 2], **kwargs)
        f.tight_layout()
        f.subplots_adjust(
            left=None, bottom=None, right=None, top=None, wspace=0.0, hspace=0.05
        )
        return f, axes

    def _plot(
        self,
        ax,
        image,
        coords,
        log_scale=True,
        v_min=None,
        v_max=None,
        text="Observed",
        font_size=15,
        colorbar_label=r"log$_{10}$ flux",
        arrow_size=0.02,
        cmap=None,
        dist_scale=1.0,
        white_on_black=True,
        no_support=False,
        **kwargs
    ):
        """

        :param ax: matplotlib axis instance
        :param image: 2d numpy array to be plotted
        :param coords: Coordinate() instance with the coordinate system
        :param white_on_black: boolean, if True, prints white text on black background, otherwise the opposite
        :return: matplotlib axis instance
        """
        if white_on_black:
            text_k = "w"
            bkg_k = "k"
        else:
            text_k = "k"
            bkg_k = "w"

        if cmap is None:
            cmap = self._cmap
        frame_size = np.max(coords.width)

        if log_scale:
            if v_min is None:
                v_min = self._v_min_default
            if v_max is None:
                v_max = self._v_max_default
            image_plot = np.log10(image)
        else:
            image_plot = image
        im = ax.matshow(
            image_plot,
            origin="lower",
            extent=[0, frame_size, 0, frame_size],
            cmap=cmap,
            vmin=v_min,
            vmax=v_max,
        )  # , vmin=0, vmax=2

        ax.get_xaxis().set_visible(False)
        ax.get_yaxis().set_visible(False)
        ax.autoscale(False)

        if not no_support:
            text_dist = "{:.1f}".format(dist_scale) + '"'
            if "no_scale_bar" not in kwargs or not kwargs["no_scale_bar"]:
                plot_util.scale_bar(
                    ax,
                    frame_size,
                    dist=dist_scale,
                    text=text_dist,
                    font_size=font_size,
                    color=text_k,
                )
            if "no_text" not in kwargs or not kwargs["no_text"]:
                plot_util.text_description(
                    ax,
                    frame_size,
                    text=text,
                    color=text_k,
                    backgroundcolor=bkg_k,
                    font_size=font_size,
                )

            if "no_arrow" not in kwargs or not kwargs["no_arrow"]:
                plot_util.coordinate_arrows(
                    ax,
                    frame_size,
                    coords,
                    color=text_k,
                    arrow_size=arrow_size,
                    font_size=font_size,
                )

            divider = make_axes_locatable(ax)
            cax = divider.append_axes("right", size="5%", pad=0.05)
            cb = plt.colorbar(im, cax=cax, orientation="vertical")
            cb.set_label(colorbar_label, fontsize=font_size)
        return ax



from lenstronomy.GalKin.aperture import Aperture

import pytest
import unittest


class TestAperture(object):
    def setup_method(self):
        pass

    def test_aperture_select(self):
        kwargs_slit = {
            "length": 2,
            "width": 0.5,
            "center_ra": 0,
            "center_dec": 0,
            "angle": 0,
        }
        slit = Aperture(aperture_type="slit", **kwargs_slit)
        bool, i = slit.aperture_select(ra=0.9, dec=0.2)
        assert bool is True
        bool, i = slit.aperture_select(ra=1.1, dec=0.2)
        assert bool is False
        assert slit.num_segments == 1

        kwargs_shell = {"r_in": 0.2, "r_out": 1.0, "center_ra": 0, "center_dec": 0}
        shell = Aperture(aperture_type="shell", **kwargs_shell)
        bool, i = shell.aperture_select(ra=0.9, dec=0)
        assert bool is True
        bool, i = shell.aperture_select(ra=1.1, dec=0)
        assert bool is False
        bool, i = shell.aperture_select(ra=0.1, dec=0)
        assert bool is False
        assert shell.num_segments == 1

        kwargs_boxhole = {
            "width_outer": 1,
            "width_inner": 0.5,
            "center_ra": 0,
            "center_dec": 0,
        }
        frame = Aperture(aperture_type="frame", **kwargs_boxhole)
        bool, i = frame.aperture_select(ra=0.4, dec=0)
        assert bool is True
        bool, i = frame.aperture_select(ra=0.2, dec=0)
        assert bool is False
        assert frame.num_segments == 1


class TestRaise(unittest.TestCase):
    def test_raise(self):
        with self.assertRaises(ValueError):
            Aperture(aperture_type="wrong", kwargs_aperture={})


if __name__ == "__main__":
    pytest.main()
